# Copyright (C) 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Implementation tests for breezy.merge.Merger."""

import os

from .. import errors
from .. import merge as _mod_merge
from ..bzr.conflicts import TextConflict
from . import TestCaseWithTransport, multiply_tests
from .test_merge_core import MergeBuilder


def load_tests(loader, standard_tests, pattern):
    """Multiply tests for tranport implementations."""
    result = loader.suiteClass()
    scenarios = [
        (name, {'merge_type': merger})
        for name, merger in _mod_merge.merge_type_registry.items()]
    return multiply_tests(standard_tests, scenarios, result)


class TestMergeImplementation(TestCaseWithTransport):

    def do_merge(self, target_tree, source_tree, **kwargs):
        merger = _mod_merge.Merger.from_revision_ids(
            target_tree, source_tree.last_revision(),
            other_branch=source_tree.branch)
        merger.merge_type = self.merge_type
        for name, value in kwargs.items():
            setattr(merger, name, value)
        merger.do_merge()

    def test_merge_specific_file(self):
        this_tree = self.make_branch_and_tree('this')
        this_tree.lock_write()
        self.addCleanup(this_tree.unlock)
        self.build_tree_contents([
            ('this/file1', b'a\nb\n'),
            ('this/file2', b'a\nb\n')
        ])
        this_tree.add(['file1', 'file2'])
        this_tree.commit('Added files')
        other_tree = this_tree.controldir.sprout('other').open_workingtree()
        self.build_tree_contents([
            ('other/file1', b'a\nb\nc\n'),
            ('other/file2', b'a\nb\nc\n')
        ])
        other_tree.commit('modified both')
        self.build_tree_contents([
            ('this/file1', b'd\na\nb\n'),
            ('this/file2', b'd\na\nb\n')
        ])
        this_tree.commit('modified both')
        self.do_merge(this_tree, other_tree, interesting_files=['file1'])
        self.assertFileEqual(b'd\na\nb\nc\n', 'this/file1')
        self.assertFileEqual(b'd\na\nb\n', 'this/file2')

    def test_merge_move_and_change(self):
        this_tree = self.make_branch_and_tree('this')
        this_tree.lock_write()
        self.addCleanup(this_tree.unlock)
        self.build_tree_contents([
            ('this/file1', b'line 1\nline 2\nline 3\nline 4\n'),
        ])
        this_tree.add('file1',)
        this_tree.commit('Added file')
        other_tree = this_tree.controldir.sprout('other').open_workingtree()
        self.build_tree_contents([
            ('other/file1', b'line 1\nline 2 to 2.1\nline 3\nline 4\n'),
        ])
        other_tree.commit('Changed 2 to 2.1')
        self.build_tree_contents([
            ('this/file1', b'line 1\nline 3\nline 2\nline 4\n'),
        ])
        this_tree.commit('Swapped 2 & 3')
        self.do_merge(this_tree, other_tree)
        if self.merge_type is _mod_merge.LCAMerger:
            self.expectFailure(
                "lca merge doesn't conflict for move and change",
                self.assertFileEqual,
                'line 1\n'
                '<<<<<<< TREE\n'
                'line 3\n'
                'line 2\n'
                '=======\n'
                'line 2 to 2.1\n'
                'line 3\n'
                '>>>>>>> MERGE-SOURCE\n'
                'line 4\n', 'this/file1')
        else:
            self.assertFileEqual(
                'line 1\n'
                '<<<<<<< TREE\n'
                'line 3\n'
                'line 2\n'
                '=======\n'
                'line 2 to 2.1\n'
                'line 3\n'
                '>>>>>>> MERGE-SOURCE\n'
                'line 4\n', 'this/file1')

    def test_modify_conflicts_with_delete(self):
        # If one side deletes a line, and the other modifies that line, then
        # the modification should be considered a conflict
        builder = self.make_branch_builder('test')
        builder.start_series()
        base_id = builder.build_snapshot(None,
                               [('add', ('', None, 'directory', None)),
                                ('add', ('foo', b'foo-id', 'file', b'a\nb\nc\nd\ne\n')),
                                ])
        # Delete 'b\n'
        other_id = builder.build_snapshot([base_id],
                               [('modify', ('foo', b'a\nc\nd\ne\n'))])
        # Modify 'b\n', add 'X\n'
        this_id = builder.build_snapshot([base_id],
                               [('modify', ('foo', b'a\nb2\nc\nd\nX\ne\n'))])
        builder.finish_series()
        branch = builder.get_branch()
        this_tree = branch.controldir.create_workingtree()
        this_tree.lock_write()
        self.addCleanup(this_tree.unlock)
        other_tree = this_tree.controldir.sprout(
            'other', other_id).open_workingtree()
        self.do_merge(this_tree, other_tree)
        if self.merge_type is _mod_merge.LCAMerger:
            self.expectFailure("lca merge doesn't track deleted lines",
                               self.assertFileEqual,
                               'a\n'
                               '<<<<<<< TREE\n'
                               'b2\n'
                               '=======\n'
                               '>>>>>>> MERGE-SOURCE\n'
                               'c\n'
                               'd\n'
                               'X\n'
                               'e\n', 'test/foo')
        else:
            self.assertFileEqual(
                b'a\n'
                b'<<<<<<< TREE\n'
                b'b2\n'
                b'=======\n'
                b'>>>>>>> MERGE-SOURCE\n'
                b'c\n'
                b'd\n'
                b'X\n'
                b'e\n', 'test/foo')

    def get_limbodir_deletiondir(self, wt):
        transform = wt.transform()
        limbodir = transform._limbodir
        deletiondir = transform._deletiondir
        transform.finalize()
        return (limbodir, deletiondir)

    def test_merge_with_existing_limbo_empty(self):
        """Empty limbo dir is just cleaned up - see bug 427773"""
        wt = self.make_branch_and_tree('this')
        (limbodir, deletiondir) = self.get_limbodir_deletiondir(wt)
        os.mkdir(limbodir)
        self.do_merge(wt, wt)

    def test_merge_with_existing_limbo_non_empty(self):
        wt = self.make_branch_and_tree('this')
        (limbodir, deletiondir) = self.get_limbodir_deletiondir(wt)
        os.mkdir(limbodir)
        os.mkdir(os.path.join(limbodir, 'something'))
        self.assertRaises(errors.ExistingLimbo, self.do_merge, wt, wt)
        self.assertRaises(errors.LockError, wt.unlock)

    def test_merge_with_pending_deletion_empty(self):
        wt = self.make_branch_and_tree('this')
        (limbodir, deletiondir) = self.get_limbodir_deletiondir(wt)
        os.mkdir(deletiondir)
        self.do_merge(wt, wt)

    def test_merge_with_pending_deletion_non_empty(self):
        """Also see bug 427773"""
        wt = self.make_branch_and_tree('this')
        (limbodir, deletiondir) = self.get_limbodir_deletiondir(wt)
        os.mkdir(deletiondir)
        os.mkdir(os.path.join(deletiondir, 'something'))
        self.assertRaises(errors.ExistingPendingDeletion,
                          self.do_merge, wt, wt)
        self.assertRaises(errors.LockError, wt.unlock)


class TestHookMergeFileContent(TestCaseWithTransport):
    """Tests that the 'merge_file_content' hook is invoked."""

    def setUp(self):
        super().setUp()
        self.hook_log = []

    def install_hook_inactive(self):
        def inactive_factory(merger):
            # This hook is never active
            self.hook_log.append(('inactive',))
            return None
        _mod_merge.Merger.hooks.install_named_hook(
            'merge_file_content', inactive_factory, 'test hook (inactive)')

    def install_hook_noop(self):
        test = self

        class HookNA(_mod_merge.AbstractPerFileMerger):
            def merge_contents(self, merge_params):
                # This hook unconditionally does nothing.
                test.hook_log.append(('no-op',))
                return 'not_applicable', None

        def hook_na_factory(merger):
            return HookNA(merger)
        _mod_merge.Merger.hooks.install_named_hook(
            'merge_file_content', hook_na_factory, 'test hook (no-op)')

    def install_hook_success(self):
        test = self

        class HookSuccess(_mod_merge.AbstractPerFileMerger):
            def merge_contents(self, merge_params):
                test.hook_log.append(('success',))
                if merge_params.this_path == 'name1':
                    return 'success', [b'text-merged-by-hook']
                return 'not_applicable', None

        def hook_success_factory(merger):
            return HookSuccess(merger)
        _mod_merge.Merger.hooks.install_named_hook(
            'merge_file_content', hook_success_factory, 'test hook (success)')

    def install_hook_conflict(self):
        test = self

        class HookConflict(_mod_merge.AbstractPerFileMerger):
            def merge_contents(self, merge_params):
                test.hook_log.append(('conflict',))
                if merge_params.this_path == 'name1':
                    return ('conflicted',
                            [b'text-with-conflict-markers-from-hook'])
                return 'not_applicable', None

        def hook_conflict_factory(merger):
            return HookConflict(merger)
        _mod_merge.Merger.hooks.install_named_hook(
            'merge_file_content', hook_conflict_factory, 'test hook (delete)')

    def install_hook_delete(self):
        test = self

        class HookDelete(_mod_merge.AbstractPerFileMerger):
            def merge_contents(self, merge_params):
                test.hook_log.append(('delete',))
                if merge_params.this_path == 'name1':
                    return 'delete', None
                return 'not_applicable', None

        def hook_delete_factory(merger):
            return HookDelete(merger)
        _mod_merge.Merger.hooks.install_named_hook(
            'merge_file_content', hook_delete_factory, 'test hook (delete)')

    def install_hook_log_lines(self):
        """Install a hook that saves the get_lines for the this, base and other
        versions of the file.
        """
        test = self

        class HookLogLines(_mod_merge.AbstractPerFileMerger):
            def merge_contents(self, merge_params):
                test.hook_log.append((
                    'log_lines',
                    merge_params.this_lines,
                    merge_params.other_lines,
                    merge_params.base_lines,
                    ))
                return 'not_applicable', None

        def hook_log_lines_factory(merger):
            return HookLogLines(merger)
        _mod_merge.Merger.hooks.install_named_hook(
            'merge_file_content', hook_log_lines_factory,
            'test hook (log_lines)')

    def make_merge_builder(self):
        builder = MergeBuilder(self.test_base_dir)
        self.addCleanup(builder.cleanup)
        return builder

    def create_file_needing_contents_merge(self, builder, name):
        file_id = name.encode('ascii') + b'-id'
        transid = builder.add_file(builder.root(), name, b"text1", True, file_id=file_id)
        builder.change_contents(transid, other=b"text4", this=b"text3")

    def test_change_vs_change(self):
        """Hook is used for (changed, changed)"""
        self.install_hook_success()
        builder = self.make_merge_builder()
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text4", this=b"text3")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual(conflicts, [])
        with builder.this.get_file('name1') as f:
            self.assertEqual(f.read(), b'text-merged-by-hook')

    def test_change_vs_deleted(self):
        """Hook is used for (changed, deleted)"""
        self.install_hook_success()
        builder = self.make_merge_builder()
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, this=b"text2")
        builder.remove_file(name1, other=True)
        conflicts = builder.merge(self.merge_type)
        self.assertEqual(conflicts, [])
        with builder.this.get_file('name1') as f:
            self.assertEqual(f.read(), b'text-merged-by-hook')

    def test_result_can_be_delete(self):
        """A hook's result can be the deletion of a file."""
        self.install_hook_delete()
        builder = self.make_merge_builder()
        self.create_file_needing_contents_merge(builder, "name1")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual(conflicts, [])
        self.assertFalse(builder.this.is_versioned('name1'))
        self.assertEqual([], list(builder.this.list_files()))

    def test_result_can_be_conflict(self):
        """A hook's result can be a conflict."""
        self.install_hook_conflict()
        builder = self.make_merge_builder()
        self.create_file_needing_contents_merge(builder, "name1")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual(1, len(conflicts))
        [conflict] = conflicts
        self.assertEqual('text conflict', conflict.typestring)
        if builder.this.supports_file_ids:
            self.assertEqual(conflict.file_id, builder.this.path2id('name1'))
        # The hook still gets to set the file contents in this case, so that it
        # can insert custom conflict markers.
        with builder.this.get_file('name1') as f:
            self.assertEqual(f.read(), b'text-with-conflict-markers-from-hook')

    def test_can_access_this_other_and_base_versions(self):
        """The hook function can call params.merger.get_lines to access the
        THIS/OTHER/BASE versions of the file.
        """
        self.install_hook_log_lines()
        builder = self.make_merge_builder()
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, this=b"text2", other=b"text3")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual(
            [('log_lines', [b'text2'], [b'text3'], [b'text1'])], self.hook_log)

    def test_chain_when_not_active(self):
        """When a hook function returns None, merging still works."""
        self.install_hook_inactive()
        self.install_hook_success()
        builder = self.make_merge_builder()
        self.create_file_needing_contents_merge(builder, "name1")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual(conflicts, [])
        with builder.this.get_file('name1') as f:
            self.assertEqual(f.read(), b'text-merged-by-hook')
        self.assertEqual([('inactive',), ('success',)], self.hook_log)

    def test_chain_when_not_applicable(self):
        """When a hook function returns not_applicable, the next function is
        tried (when one exists).
        """
        self.install_hook_noop()
        self.install_hook_success()
        builder = self.make_merge_builder()
        self.create_file_needing_contents_merge(builder, "name1")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual(conflicts, [])
        with builder.this.get_file('name1') as f:
            self.assertEqual(f.read(), b'text-merged-by-hook')
        self.assertEqual([('no-op',), ('success',)], self.hook_log)

    def test_chain_stops_after_success(self):
        """When a hook function returns success, no later functions are tried.
        """
        self.install_hook_success()
        self.install_hook_noop()
        builder = self.make_merge_builder()
        self.create_file_needing_contents_merge(builder, "name1")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual([('success',)], self.hook_log)

    def test_chain_stops_after_conflict(self):
        """When a hook function returns conflict, no later functions are tried.
        """
        self.install_hook_conflict()
        self.install_hook_noop()
        builder = self.make_merge_builder()
        self.create_file_needing_contents_merge(builder, "name1")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual([('conflict',)], self.hook_log)

    def test_chain_stops_after_delete(self):
        """When a hook function returns delete, no later functions are tried.
        """
        self.install_hook_delete()
        self.install_hook_noop()
        builder = self.make_merge_builder()
        self.create_file_needing_contents_merge(builder, "name1")
        conflicts = builder.merge(self.merge_type)
        self.assertEqual([('delete',)], self.hook_log)



# Copyright (C) 2005, 2006, 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A store that keeps the full text of every version.

This store keeps uncompressed versions of the full text. It does not
do any sort of delta compression.
"""

import gzip
import os
from io import BytesIO

from .... import osutils
from ....errors import BzrError
from ....trace import mutter
from ....transport import FileExists, NoSuchFile
from . import TransportStore


class TextStore(TransportStore):
    """Store that holds files indexed by unique names.

    Files can be added, but not modified once they are in.  Typically
    the hash is used as the name, or something else known to be unique,
    such as a UUID.

    Files are stored uncompressed, with no delta compression.
    """

    def _add_compressed(self, fn, f):
        if isinstance(f, bytes):
            f = BytesIO(f)

        sio = BytesIO()
        gf = gzip.GzipFile(mode='wb', fileobj=sio)
        # if pumpfile handles files that don't fit in ram,
        # so will this function
        osutils.pumpfile(f, gf)
        gf.close()
        sio.seek(0)
        self._try_put(fn, sio)

    def _add(self, fn, f):
        if self._compressed:
            self._add_compressed(fn, f)
        else:
            self._try_put(fn, f)

    def _try_put(self, fn, f):
        try:
            self._transport.put_file(fn, f, mode=self._file_mode)
        except NoSuchFile:
            if not self._prefixed:
                raise
            try:
                self._transport.mkdir(os.path.dirname(fn), mode=self._dir_mode)
            except FileExists:
                pass
            self._transport.put_file(fn, f, mode=self._file_mode)

    def _get(self, fn):
        if fn.endswith('.gz'):
            return self._get_compressed(fn)
        else:
            return self._transport.get(fn)

    def _get_compressed(self, filename):
        """Returns a file reading from a particular entry."""
        f = self._transport.get(filename)
        # gzip.GzipFile.read() requires a tell() function
        # but some transports return objects that cannot seek
        # so buffer them in a BytesIO instead
        if getattr(f, 'tell', None) is not None:
            return gzip.GzipFile(mode='rb', fileobj=f)
        try:
            sio = BytesIO(f.read())
            return gzip.GzipFile(mode='rb', fileobj=sio)
        finally:
            f.close()



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for checking of trees."""

from breezy.bzr.workingtree import InventoryWorkingTree
from breezy.tests import TestNotApplicable
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestCheck(TestCaseWithWorkingTree):

    def test__get_check_refs_new(self):
        tree = self.make_branch_and_tree('tree')
        if not isinstance(tree, InventoryWorkingTree):
            raise TestNotApplicable(
                "_get_check_refs only relevant for inventory working trees")
        self.assertEqual({('trees', b'null:')},
                         set(tree._get_check_refs()))

    def test__get_check_refs_basis(self):
        # with a basis, all current bzr trees cache it and so need the
        # inventory to cross-check.
        tree = self.make_branch_and_tree('tree')
        if not isinstance(tree, InventoryWorkingTree):
            raise TestNotApplicable(
                "_get_check_refs only relevant for inventory working trees")
        revid = tree.commit('first post')
        self.assertEqual({('trees', revid)},
                         set(tree._get_check_refs()))

    def test__check_with_refs(self):
        # _check can be called with a dict of the things required.
        tree = self.make_branch_and_tree('tree')
        if not isinstance(tree, InventoryWorkingTree):
            raise TestNotApplicable(
                "_get_check_refs only relevant for inventory working trees")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        revid = tree.commit('first post')
        needed_refs = tree._get_check_refs()
        repo = tree.branch.repository
        for ref in needed_refs:
            kind, revid = ref
            refs = {}
            if kind == 'trees':
                refs[ref] = repo.revision_tree(revid)
            else:
                self.fail('unknown ref kind')
        tree._check(refs)



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'WorkingTree.annotate_iter'"""

from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestAnnotateIter(TestCaseWithWorkingTree):

    def make_single_rev_tree(self):
        builder = self.make_branch_builder('branch')
        revid = builder.build_snapshot(None, [
            ('add', ('', None, 'directory', None)),
            ('add', ('file', None, 'file', b'initial content\n')),
            ])
        b = builder.get_branch()
        tree = b.create_checkout('tree', lightweight=True)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        return tree, revid

    def test_annotate_same_as_parent(self):
        tree, revid = self.make_single_rev_tree()
        annotations = tree.annotate_iter('file')
        self.assertEqual([(revid, b'initial content\n')],
                         annotations)

    def test_annotate_mod_from_parent(self):
        tree, revid = self.make_single_rev_tree()
        self.build_tree_contents([('tree/file',
                                   b'initial content\nnew content\n')])
        annotations = tree.annotate_iter('file')
        self.assertEqual([(revid, b'initial content\n'),
                          (b'current:', b'new content\n'),
                          ], annotations)

    def test_annotate_merge_parents(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        revid1 = builder.build_snapshot(None, [
            ('add', ('', None, 'directory', None)),
            ('add', ('file', None, 'file', b'initial content\n')),
            ])
        revid2 = builder.build_snapshot([revid1], [
            ('modify', ('file', b'initial content\ncontent in 2\n')),
            ])
        revid3 = builder.build_snapshot([revid1], [
            ('modify', ('file', b'initial content\ncontent in 3\n')),
            ])
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout('tree', revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents([('tree/file',
                                   b'initial content\ncontent in 2\n'
                                   b'content in 3\nnew content\n')])
        annotations = tree.annotate_iter('file')
        self.assertEqual([(revid1, b'initial content\n'),
                          (revid2, b'content in 2\n'),
                          (revid3, b'content in 3\n'),
                          (b'current:', b'new content\n'),
                          ], annotations)

    def test_annotate_merge_parent_no_file(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        revid1 = builder.build_snapshot(None, [
            ('add', ('', None, 'directory', None)),
            ])
        revid2 = builder.build_snapshot([revid1], [
            ('add', ('file', None, 'file', b'initial content\n')),
            ])
        revid3 = builder.build_snapshot([revid1], [])
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout('tree', revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents([('tree/file',
                                   b'initial content\nnew content\n')])
        annotations = tree.annotate_iter('file')
        self.assertEqual([(revid2, b'initial content\n'),
                          (b'current:', b'new content\n'),
                          ], annotations)

    def test_annotate_merge_parent_was_directory(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        revid1 = builder.build_snapshot(None, [
            ('add', ('', None, 'directory', None)),
            ])
        revid2 = builder.build_snapshot([revid1], [
            ('add', ('file', None, 'file', b'initial content\n')),
            ])
        revid3 = builder.build_snapshot([revid1], [
            ('add', ('a_dir', None, 'directory', None)),
            ])
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout('tree', revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents([('tree/file',
                                   b'initial content\nnew content\n')])
        annotations = tree.annotate_iter('file')
        self.assertEqual([(revid2, b'initial content\n'),
                          (b'current:', b'new content\n'),
                          ], annotations)

    def test_annotate_same_as_merge_parent(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        revid1 = builder.build_snapshot(None, [
            ('add', ('', None, 'directory', None)),
            ('add', ('file', None, 'file', b'initial content\n')),
            ])
        revid2 = builder.build_snapshot([revid1], [
            ])
        revid3 = builder.build_snapshot([revid1], [
            ('modify', ('file', b'initial content\ncontent in 3\n')),
            ])
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout('tree', revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents([('tree/file',
                                   b'initial content\ncontent in 3\n')])
        annotations = tree.annotate_iter('file')
        self.assertEqual([(revid1, b'initial content\n'),
                          (revid3, b'content in 3\n'),
                          ], annotations)

    def test_annotate_same_as_merge_parent_supersedes(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        revid1 = builder.build_snapshot(None, [
            ('add', ('', None, 'directory', None)),
            ('add', ('file', None, 'file', b'initial content\n')),
            ])
        revid2 = builder.build_snapshot([revid1], [
            ('modify', ('file', b'initial content\nnew content\n')),
            ])
        revid3 = builder.build_snapshot([revid2], [
            ('modify', ('file', b'initial content\ncontent in 3\n')),
            ])
        revid4 = builder.build_snapshot([revid3], [
            ('modify', ('file', b'initial content\nnew content\n')),
            ])
        # In this case, the content locally is the same as content in basis
        # tree, but the merge revision states that *it* should win
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout('tree', revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid4])
        annotations = tree.annotate_iter('file')
        self.assertEqual([(revid1, b'initial content\n'),
                          (revid4, b'new content\n'),
                          ], annotations)



# Copyright (C) 2007 Canonical Ltd
# Copyright (C) 2007-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interfacing with a Git Repository"""

import os

import dulwich
from dulwich.repo import Repo as GitRepo

from ... import config, errors, revision
from ...repository import InterRepository, Repository
from .. import dir, repository, tests
from ..mapping import default_mapping
from ..object_store import BazaarObjectStore
from ..push import MissingObjectsIterator


class TestGitRepositoryFeatures(tests.TestCaseInTempDir):
    """Feature tests for GitRepository."""

    def _do_commit(self):
        builder = tests.GitBranchBuilder()
        builder.set_file(b'a', b'text for a\n', False)
        commit_handle = builder.commit(b'Joe Foo <joe@foo.com>', b'message')
        mapping = builder.finish()
        return mapping[commit_handle]

    def test_open_existing(self):
        GitRepo.init(self.test_dir)

        repo = Repository.open('.')
        self.assertIsInstance(repo, repository.GitRepository)

    def test_has_git_repo(self):
        GitRepo.init(self.test_dir)

        repo = Repository.open('.')
        self.assertIsInstance(repo._git, dulwich.repo.BaseRepo)

    def test_has_revision(self):
        GitRepo.init(self.test_dir)
        commit_id = self._do_commit()
        repo = Repository.open('.')
        self.assertFalse(repo.has_revision(b'foobar'))
        revid = default_mapping.revision_id_foreign_to_bzr(commit_id)
        self.assertTrue(repo.has_revision(revid))

    def test_has_revisions(self):
        GitRepo.init(self.test_dir)
        commit_id = self._do_commit()
        repo = Repository.open('.')
        self.assertEqual(set(), repo.has_revisions([b'foobar']))
        revid = default_mapping.revision_id_foreign_to_bzr(commit_id)
        self.assertEqual({revid}, repo.has_revisions([b'foobar', revid]))

    def test_get_revision(self):
        # GitRepository.get_revision gives a Revision object.

        # Create a git repository with a revision.
        GitRepo.init(self.test_dir)
        commit_id = self._do_commit()

        # Get the corresponding Revision object.
        revid = default_mapping.revision_id_foreign_to_bzr(commit_id)
        repo = Repository.open('.')
        rev = repo.get_revision(revid)
        self.assertIsInstance(rev, revision.Revision)

    def test_get_revision_unknown(self):
        GitRepo.init(self.test_dir)

        repo = Repository.open('.')
        self.assertRaises(errors.NoSuchRevision, repo.get_revision, b"bla")

    def simple_commit(self):
        # Create a git repository with some interesting files in a revision.
        GitRepo.init(self.test_dir)
        builder = tests.GitBranchBuilder()
        builder.set_file(b'data', b'text\n', False)
        builder.set_file(b'executable', b'content', True)
        builder.set_symlink(b'link', b'broken')
        builder.set_file(b'subdir/subfile', b'subdir text\n', False)
        commit_handle = builder.commit(b'Joe Foo <joe@foo.com>', b'message',
                                       timestamp=1205433193)
        mapping = builder.finish()
        return mapping[commit_handle]

    def test_pack(self):
        commit_id = self.simple_commit()
        repo = Repository.open('.')
        repo.pack()

    def test_unlock_closes(self):
        commit_id = self.simple_commit()
        repo = Repository.open('.')
        repo.pack()
        with repo.lock_read():
            repo.all_revision_ids()
            self.assertTrue(len(repo._git.object_store._pack_cache) > 0)
        self.assertEqual(len(repo._git.object_store._pack_cache), 0)

    def test_revision_tree(self):
        commit_id = self.simple_commit()
        revid = default_mapping.revision_id_foreign_to_bzr(commit_id)
        repo = Repository.open('.')
        tree = repo.revision_tree(revid)
        self.assertEqual(tree.get_revision_id(), revid)
        self.assertEqual(b"text\n", tree.get_file_text("data"))


class TestGitRepository(tests.TestCaseWithTransport):

    def _do_commit(self):
        builder = tests.GitBranchBuilder()
        builder.set_file(b'a', b'text for a\n', False)
        commit_handle = builder.commit(b'Joe Foo <joe@foo.com>', b'message')
        mapping = builder.finish()
        return mapping[commit_handle]

    def setUp(self):
        tests.TestCaseWithTransport.setUp(self)
        dulwich.repo.Repo.create(self.test_dir)
        self.git_repo = Repository.open(self.test_dir)

    def test_supports_rich_root(self):
        repo = self.git_repo
        self.assertEqual(repo.supports_rich_root(), True)

    def test_get_signature_text(self):
        self.assertRaises(
            errors.NoSuchRevision, self.git_repo.get_signature_text, revision.NULL_REVISION)

    def test_has_signature_for_revision_id(self):
        self.assertEqual(False, self.git_repo.has_signature_for_revision_id(
            revision.NULL_REVISION))

    def test_all_revision_ids_none(self):
        self.assertEqual([], self.git_repo.all_revision_ids())

    def test_get_known_graph_ancestry(self):
        cid = self._do_commit()
        revid = default_mapping.revision_id_foreign_to_bzr(cid)
        g = self.git_repo.get_known_graph_ancestry([revid])
        self.assertEqual(frozenset([revid]),
                         g.heads([revid]))
        self.assertEqual([(revid, 0, (1,), True)],
                         [(n.key, n.merge_depth, n.revno, n.end_of_merge)
                          for n in g.merge_sort(revid)])

    def test_all_revision_ids(self):
        commit_id = self._do_commit()
        self.assertEqual(
            [default_mapping.revision_id_foreign_to_bzr(commit_id)],
            self.git_repo.all_revision_ids())

    def assertIsNullInventory(self, inv):
        self.assertEqual(inv.root, None)
        self.assertEqual(inv.revision_id, revision.NULL_REVISION)
        self.assertEqual(list(inv.iter_entries()), [])

    def test_revision_tree_none(self):
        # GitRepository.revision_tree('null':') returns the null tree.
        repo = self.git_repo
        tree = repo.revision_tree(revision.NULL_REVISION)
        self.assertEqual(tree.get_revision_id(), revision.NULL_REVISION)

    def test_get_parent_map_null(self):
        self.assertEqual({revision.NULL_REVISION: ()},
                         self.git_repo.get_parent_map([revision.NULL_REVISION]))


class SigningGitRepository(tests.TestCaseWithTransport):

    def test_signed_commit(self):
        import breezy.gpg
        oldstrategy = breezy.gpg.GPGStrategy
        wt = self.make_branch_and_tree('.', format='git')
        branch = wt.branch
        revid = wt.commit("base", allow_pointless=True)
        self.assertFalse(
            branch.repository.has_signature_for_revision_id(revid))
        try:
            breezy.gpg.GPGStrategy = breezy.gpg.LoopbackGPGStrategy
            conf = config.MemoryStack(b'''
create_signatures=always
''')
            revid2 = wt.commit(config=conf, message="base",
                               allow_pointless=True)

            def sign(text):
                return breezy.gpg.LoopbackGPGStrategy(None).sign(text)
            self.assertIsInstance(
                branch.repository.get_signature_text(revid2), bytes)
        finally:
            breezy.gpg.GPGStrategy = oldstrategy


class RevpropsRepository(tests.TestCaseWithTransport):

    def test_author(self):
        wt = self.make_branch_and_tree('.', format='git')
        revid = wt.commit(
            "base", allow_pointless=True,
            revprops={'author': 'Joe Example <joe@example.com>'})
        rev = wt.branch.repository.get_revision(revid)
        r = dulwich.repo.Repo('.')
        self.assertEqual(b'Joe Example <joe@example.com>', r[r.head()].author)

    def test_authors_single_author(self):
        wt = self.make_branch_and_tree('.', format='git')
        revid = wt.commit(
            "base", allow_pointless=True,
            revprops={'authors': 'Joe Example <joe@example.com>'})
        rev = wt.branch.repository.get_revision(revid)
        r = dulwich.repo.Repo('.')
        self.assertEqual(b'Joe Example <joe@example.com>', r[r.head()].author)

    def test_multiple_authors(self):
        wt = self.make_branch_and_tree('.', format='git')
        self.assertRaises(
            Exception, wt.commit, "base", allow_pointless=True,
            revprops={'authors': 'Joe Example <joe@example.com>\n'
                                 'Jane Doe <jane@example.com\n>'})

    def test_bugs(self):
        wt = self.make_branch_and_tree('.', format='git')
        revid = wt.commit(
            "base", allow_pointless=True,
            revprops={
                'bugs': 'https://github.com/jelmer/dulwich/issues/123 fixed\n'
                })
        rev = wt.branch.repository.get_revision(revid)
        r = dulwich.repo.Repo('.')
        self.assertEqual(
            b'base\n\n'
            b'Fixes: https://github.com/jelmer/dulwich/issues/123\n',
            r[r.head()].message)

    def test_authors(self):
        wt = self.make_branch_and_tree('.', format='git')
        revid = wt.commit(
            "base", allow_pointless=True,
            revprops={
                'authors': (
                    'Jelmer Vernooij <jelmer@example.com>\n'
                    'Martin Packman <bz2@example.com>\n'),
                })
        rev = wt.branch.repository.get_revision(revid)
        r = dulwich.repo.Repo('.')
        self.assertEqual(
            r[r.head()].author, b'Jelmer Vernooij <jelmer@example.com>')
        self.assertEqual(
            b'base\n\nCo-authored-by: Martin Packman <bz2@example.com>\n',
            r[r.head()].message)


class GitRepositoryFormat(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.format = repository.GitRepositoryFormat()

    def test_get_format_description(self):
        self.assertEqual("Git Repository",
                         self.format.get_format_description())


class RevisionGistImportTests(tests.TestCaseWithTransport):

    def setUp(self):
        tests.TestCaseWithTransport.setUp(self)
        self.git_path = os.path.join(self.test_dir, "git")
        os.mkdir(self.git_path)
        dulwich.repo.Repo.create(self.git_path)
        self.git_repo = Repository.open(self.git_path)
        self.bzr_tree = self.make_branch_and_tree("bzr")

    def get_inter(self):
        return InterRepository.get(self.bzr_tree.branch.repository,
                                   self.git_repo)

    def object_iter(self):
        store = BazaarObjectStore(
            self.bzr_tree.branch.repository, default_mapping)
        store_iterator = MissingObjectsIterator(
            store, self.bzr_tree.branch.repository)
        return store, store_iterator

    def import_rev(self, revid, parent_lookup=None):
        store, store_iter = self.object_iter()
        store._cache.idmap.start_write_group()
        try:
            return store_iter.import_revision(revid, lossy=True)
        except:
            store._cache.idmap.abort_write_group()
            raise
        else:
            store._cache.idmap.commit_write_group()

    def test_pointless(self):
        revid = self.bzr_tree.commit("pointless", timestamp=1205433193,
                                     timezone=0, committer="Jelmer Vernooij <jelmer@samba.org>")
        self.assertEqual(b"2caa8094a5b794961cd9bf582e3e2bb090db0b14",
                         self.import_rev(revid))
        self.assertEqual(b"2caa8094a5b794961cd9bf582e3e2bb090db0b14",
                         self.import_rev(revid))


class ForeignTestsRepositoryFactory:

    def make_repository(self, transport):
        return dir.LocalGitControlDirFormat().initialize_on_transport(transport).open_repository()



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for get_rev_id_for_revno."""

from breezy import errors
from breezy.tests.per_repository_reference import \
    TestCaseWithExternalReferenceRepository


class TestGetRevIdForRevno(TestCaseWithExternalReferenceRepository):

    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree('base')
        self.revid1 = self.tree.commit('one')
        self.revid2 = self.tree.commit('two')
        self.revid3 = self.tree.commit('three')

    def test_success(self):
        repo = self.tree.branch.repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual(
            (True, self.revid1),
            repo.get_rev_id_for_revno(1, (3, self.revid3)))
        self.assertEqual(
            (True, self.revid2),
            repo.get_rev_id_for_revno(2, (3, self.revid3)))

    def test_unknown_revision(self):
        tree2 = self.make_branch_and_tree('other')
        unknown_revid = tree2.commit('other')
        repo = self.tree.branch.repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertRaises(
            errors.NoSuchRevision,
            repo.get_rev_id_for_revno, 1, (3, unknown_revid))

    def test_known_pair_is_after(self):
        repo = self.tree.branch.repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertRaises(
            errors.RevnoOutOfBounds,
            repo.get_rev_id_for_revno, 3, (2, self.revid2))



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ... import graph as _mod_graph
from ... import tests
from ...revision import NULL_REVISION
from ...tests.test_graph import TestGraphBase
from .. import vf_search

# Ancestry 1:
#
#  NULL_REVISION
#       |
#     rev1
#      /\
#  rev2a rev2b
#     |    |
#   rev3  /
#     |  /
#   rev4
ancestry_1 = {b'rev1': [NULL_REVISION],
              b'rev2a': [b'rev1'],
              b'rev2b': [b'rev1'],
              b'rev3': [b'rev2a'],
              b'rev4': [b'rev3', b'rev2b']}

# Ancestry 2:
#
#  NULL_REVISION
#    /    \
# rev1a  rev1b
#   |
# rev2a
#   |
# rev3a
#   |
# rev4a
ancestry_2 = {b'rev1a': [NULL_REVISION],
              b'rev2a': [b'rev1a'],
              b'rev1b': [NULL_REVISION],
              b'rev3a': [b'rev2a'],
              b'rev4a': [b'rev3a']}


# Extended history shortcut
#  NULL_REVISION
#       |
#       a
#       |\
#       b |
#       | |
#       c |
#       | |
#       d |
#       |\|
#       e f
extended_history_shortcut = {b'a': [NULL_REVISION],
                             b'b': [b'a'],
                             b'c': [b'b'],
                             b'd': [b'c'],
                             b'e': [b'd'],
                             b'f': [b'a', b'd'],
                             }


class TestSearchResultRefine(tests.TestCase):

    def make_graph(self, ancestors):
        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))

    def test_refine(self):
        # Used when pulling from a stacked repository, so test some revisions
        # being satisfied from the stacking branch.
        self.make_graph(
            {b"tip": [b"mid"], b"mid": [b"base"], b"tag": [b"base"],
             b"base": [NULL_REVISION], NULL_REVISION: []})
        result = vf_search.SearchResult(
            {b'tip', b'tag'},
            {NULL_REVISION}, 4, {b'tip', b'mid', b'tag', b'base'})
        result = result.refine({b'tip'}, {b'mid'})
        recipe = result.get_recipe()
        # We should be starting from tag (original head) and mid (seen ref)
        self.assertEqual({b'mid', b'tag'}, recipe[1])
        # We should be stopping at NULL (original stop) and tip (seen head)
        self.assertEqual({NULL_REVISION, b'tip'}, recipe[2])
        self.assertEqual(3, recipe[3])
        result = result.refine({b'mid', b'tag', b'base'},
                               {NULL_REVISION})
        recipe = result.get_recipe()
        # We should be starting from nothing (NULL was known as a cut point)
        self.assertEqual(set(), recipe[1])
        # We should be stopping at NULL (original stop) and tip (seen head) and
        # tag (seen head) and mid(seen mid-point head). We could come back and
        # define this as not including mid, for minimal results, but it is
        # still 'correct' to include mid, and simpler/easier.
        self.assertEqual({NULL_REVISION, b'tip', b'tag', b'mid'}, recipe[2])
        self.assertEqual(0, recipe[3])
        self.assertTrue(result.is_empty())


class TestSearchResultFromParentMap(TestGraphBase):

    def assertSearchResult(self, start_keys, stop_keys, key_count, parent_map,
                           missing_keys=()):
        (start, stop, count) = vf_search.search_result_from_parent_map(
            parent_map, missing_keys)
        self.assertEqual((sorted(start_keys), sorted(stop_keys), key_count),
                         (sorted(start), sorted(stop), count))

    def test_no_parents(self):
        self.assertSearchResult([], [], 0, {})
        self.assertSearchResult([], [], 0, None)

    def test_ancestry_1(self):
        self.assertSearchResult([b'rev4'], [NULL_REVISION], len(ancestry_1),
                                ancestry_1)

    def test_ancestry_2(self):
        self.assertSearchResult([b'rev1b', b'rev4a'], [NULL_REVISION],
                                len(ancestry_2), ancestry_2)
        self.assertSearchResult([b'rev1b', b'rev4a'], [],
                                len(ancestry_2) + 1, ancestry_2,
                                missing_keys=[NULL_REVISION])

    def test_partial_search(self):
        parent_map = {k: extended_history_shortcut[k]
                          for k in [b'e', b'f']}
        self.assertSearchResult([b'e', b'f'], [b'd', b'a'], 2,
                                parent_map)
        parent_map.update((k, extended_history_shortcut[k])
                          for k in [b'd', b'a'])
        self.assertSearchResult([b'e', b'f'], [b'c', NULL_REVISION], 4,
                                parent_map)
        parent_map[b'c'] = extended_history_shortcut[b'c']
        self.assertSearchResult([b'e', b'f'], [b'b'], 6,
                                parent_map, missing_keys=[NULL_REVISION])
        parent_map[b'b'] = extended_history_shortcut[b'b']
        self.assertSearchResult([b'e', b'f'], [], 7,
                                parent_map, missing_keys=[NULL_REVISION])


class TestLimitedSearchResultFromParentMap(TestGraphBase):

    def assertSearchResult(self, start_keys, stop_keys, key_count, parent_map,
                           missing_keys, tip_keys, depth):
        (start, stop, count) = vf_search.limited_search_result_from_parent_map(
            parent_map, missing_keys, tip_keys, depth)
        self.assertEqual((sorted(start_keys), sorted(stop_keys), key_count),
                         (sorted(start), sorted(stop), count))

    def test_empty_ancestry(self):
        self.assertSearchResult([], [], 0, {}, (), [b'tip-rev-id'], 10)

    def test_ancestry_1(self):
        self.assertSearchResult([b'rev4'], [b'rev1'], 4,
                                ancestry_1, (), [b'rev1'], 10)
        self.assertSearchResult([b'rev2a', b'rev2b'], [b'rev1'], 2,
                                ancestry_1, (), [b'rev1'], 1)

    def test_multiple_heads(self):
        self.assertSearchResult([b'e', b'f'], [b'a'], 5,
                                extended_history_shortcut, (), [b'a'], 10)
        # Note that even though we only take 1 step back, we find 'f', which
        # means the described search will still find d and c.
        self.assertSearchResult([b'f'], [b'a'], 4,
                                extended_history_shortcut, (), [b'a'], 1)
        self.assertSearchResult([b'f'], [b'a'], 4,
                                extended_history_shortcut, (), [b'a'], 2)


class TestPendingAncestryResultRefine(tests.TestCase):

    def make_graph(self, ancestors):
        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))

    def test_refine(self):
        # Used when pulling from a stacked repository, so test some revisions
        # being satisfied from the stacking branch.
        g = self.make_graph(
            {b"tip": [b"mid"], b"mid": [b"base"], b"tag": [b"base"],
             b"base": [NULL_REVISION], NULL_REVISION: []})
        result = vf_search.PendingAncestryResult([b'tip', b'tag'], None)
        result = result.refine({b'tip'}, {b'mid'})
        self.assertEqual({b'mid', b'tag'}, result.heads)
        result = result.refine({b'mid', b'tag', b'base'},
                               {NULL_REVISION})
        self.assertEqual({NULL_REVISION}, result.heads)
        self.assertTrue(result.is_empty())


class TestPendingAncestryResultGetKeys(tests.TestCaseWithMemoryTransport):
    """Tests for breezy.graph.PendingAncestryResult."""

    def test_get_keys(self):
        builder = self.make_branch_builder('b')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', ''))],
            revision_id=b'rev-1')
        builder.build_snapshot([b'rev-1'], [], revision_id=b'rev-2')
        builder.finish_series()
        repo = builder.get_branch().repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        result = vf_search.PendingAncestryResult([b'rev-2'], repo)
        self.assertEqual({b'rev-1', b'rev-2'}, set(result.get_keys()))

    def test_get_keys_excludes_ghosts(self):
        builder = self.make_branch_builder('b')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', ''))],
            revision_id=b'rev-1')
        builder.build_snapshot([b'rev-1', b'ghost'], [], revision_id=b'rev-2')
        builder.finish_series()
        repo = builder.get_branch().repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        result = vf_search.PendingAncestryResult([b'rev-2'], repo)
        self.assertEqual(sorted([b'rev-1', b'rev-2']),
                         sorted(result.get_keys()))

    def test_get_keys_excludes_null(self):
        # Make a 'graph' with an iter_ancestry that returns NULL_REVISION
        # somewhere other than the last element, which can happen in real
        # ancestries.
        class StubGraph:
            def iter_ancestry(self, keys):
                return [(NULL_REVISION, ()), (b'foo', (NULL_REVISION,))]
        result = vf_search.PendingAncestryResult([b'rev-3'], None)
        result_keys = result._get_keys(StubGraph())
        # Only the non-null keys from the ancestry appear.
        self.assertEqual({b'foo'}, set(result_keys))



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Conversion between refs and Bazaar revision pointers."""

from dulwich.objects import Tag, object_class
from dulwich.refs import LOCAL_BRANCH_PREFIX, LOCAL_TAG_PREFIX

try:
    from dulwich.refs import PEELED_TAG_SUFFIX
except ImportError:
    from dulwich.refs import ANNOTATED_TAG_SUFFIX as PEELED_TAG_SUFFIX

from dulwich.repo import RefsContainer

from .. import controldir, errors, osutils
from .. import revision as _mod_revision


def is_tag(x):
    return x.startswith(LOCAL_TAG_PREFIX)


def is_peeled(x):
    return x.endswith(PEELED_TAG_SUFFIX)


def branch_name_to_ref(name):
    """Map a branch name to a ref.

    :param name: Branch name
    :return: ref string
    """
    if name == "":
        return b"HEAD"
    if not name.startswith("refs/"):
        return LOCAL_BRANCH_PREFIX + osutils.safe_utf8(name)
    else:
        return osutils.safe_utf8(name)


def tag_name_to_ref(name):
    """Map a tag name to a ref.

    :param name: Tag name
    :return: ref string
    """
    return LOCAL_TAG_PREFIX + osutils.safe_utf8(name)


def ref_to_branch_name(ref):
    """Map a ref to a branch name

    :param ref: Ref
    :return: A branch name
    """
    if ref == b"HEAD":
        return ""
    if ref is None:
        return ref
    if ref.startswith(LOCAL_BRANCH_PREFIX):
        return ref[len(LOCAL_BRANCH_PREFIX):].decode('utf-8')
    raise ValueError("unable to map ref %s back to branch name" % ref)


def ref_to_tag_name(ref):
    if ref.startswith(LOCAL_TAG_PREFIX):
        return ref[len(LOCAL_TAG_PREFIX):].decode("utf-8")
    raise ValueError("unable to map ref %s back to tag name" % ref)


class BazaarRefsContainer(RefsContainer):

    def __init__(self, dir, object_store):
        self.dir = dir
        self.object_store = object_store

    def get_packed_refs(self):
        return {}

    def set_symbolic_ref(self, name, other):
        if name == b"HEAD":
            pass  # FIXME: Switch default branch
        else:
            raise NotImplementedError(
                "Symbolic references not supported for anything other than "
                "HEAD")

    def _get_revid_by_tag_name(self, tag_name):
        for branch in self.dir.list_branches():
            try:
                # FIXME: This is ambiguous!
                return branch.tags.lookup_tag(tag_name)
            except errors.NoSuchTag:
                pass
        return None

    def _get_revid_by_branch_name(self, branch_name):
        try:
            branch = self.dir.open_branch(branch_name)
        except controldir.NoColocatedBranchSupport:
            if branch_name in ("HEAD", "master"):
                branch = self.dir.open_branch()
            else:
                raise
        return branch.last_revision()

    def read_loose_ref(self, ref):
        try:
            branch_name = ref_to_branch_name(ref)
        except ValueError:
            tag_name = ref_to_tag_name(ref)
            revid = self._get_revid_by_tag_name(tag_name)
        else:
            revid = self._get_revid_by_branch_name(branch_name)
        if revid == _mod_revision.NULL_REVISION:
            return None
        # FIXME: Unpeel if necessary
        with self.object_store.lock_read():
            return self.object_store._lookup_revision_sha1(revid)

    def get_peeled(self, ref):
        return self.read_loose_ref(ref)

    def allkeys(self):
        keys = set()
        for branch in self.dir.list_branches():
            repo = branch.repository
            if repo.has_revision(branch.last_revision()):
                ref = branch_name_to_ref(getattr(branch, "name", ""))
                keys.add(ref)
            try:
                for tag_name, revid in branch.tags.get_tag_dict().items():
                    if repo.has_revision(revid):
                        keys.add(tag_name_to_ref(tag_name))
            except errors.TagsNotSupported:
                pass
        return keys

    def __delitem__(self, ref):
        try:
            branch_name = ref_to_branch_name(ref)
        except ValueError:
            return  # FIXME: Cope with tags!
        self.dir.destroy_branch(branch_name)

    def __setitem__(self, ref, sha):
        try:
            branch_name = ref_to_branch_name(ref)
        except ValueError:
            # FIXME: Cope with tags!
            return
        try:
            target_branch = self.repo_dir.open_branch(branch_name)
        except errors.NotBranchError:
            target_branch = self.repo.create_branch(branch_name)

        rev_id = self.mapping.revision_id_foreign_to_bzr(sha)
        with target_branch.lock_write():
            target_branch.generate_revision_history(rev_id)


def get_refs_container(controldir, object_store):
    fn = getattr(controldir, "get_refs_container", None)
    if fn is not None:
        return fn()
    return BazaarRefsContainer(controldir, object_store)



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tests of the 'brz dump-btree' command."""

from breezy import tests
from breezy.bzr import btree_index
from breezy.tests import http_server


class TestDumpBtree(tests.TestCaseWithTransport):

    def create_sample_btree_index(self):
        builder = btree_index.BTreeBuilder(
            reference_lists=1, key_elements=2)
        builder.add_node((b'test', b'key1'), b'value',
                         (((b'ref', b'entry'),),))
        builder.add_node((b'test', b'key2'), b'value2',
                         (((b'ref', b'entry2'),),))
        builder.add_node((b'test2', b'key3'), b'value3',
                         (((b'ref', b'entry3'),),))
        out_f = builder.finish()
        try:
            self.build_tree_contents([('test.btree', out_f.read())])
        finally:
            out_f.close()

    def test_dump_btree_smoke(self):
        self.create_sample_btree_index()
        out, err = self.run_bzr('dump-btree test.btree')
        self.assertEqualDiff(
            "(('test', 'key1'), 'value', ((('ref', 'entry'),),))\n"
            "(('test', 'key2'), 'value2', ((('ref', 'entry2'),),))\n"
            "(('test2', 'key3'), 'value3', ((('ref', 'entry3'),),))\n",
            out)

    def test_dump_btree_http_smoke(self):
        self.transport_readonly_server = http_server.HttpServer
        self.create_sample_btree_index()
        url = self.get_readonly_url('test.btree')
        out, err = self.run_bzr(['dump-btree', url])
        self.assertEqualDiff(
            "(('test', 'key1'), 'value', ((('ref', 'entry'),),))\n"
            "(('test', 'key2'), 'value2', ((('ref', 'entry2'),),))\n"
            "(('test2', 'key3'), 'value3', ((('ref', 'entry3'),),))\n",
            out)

    def test_dump_btree_raw_smoke(self):
        self.create_sample_btree_index()
        out, err = self.run_bzr('dump-btree test.btree --raw')
        self.assertEqualDiff(
            'Root node:\n'
            'B+Tree Graph Index 2\n'
            'node_ref_lists=1\n'
            'key_elements=2\n'
            'len=3\n'
            'row_lengths=1\n'
            '\n'
            'Page 0\n'
            'type=leaf\n'
            'test\0key1\0ref\0entry\0value\n'
            'test\0key2\0ref\0entry2\0value2\n'
            'test2\0key3\0ref\0entry3\0value3\n'
            '\n',
            out)

    def test_dump_btree_no_refs_smoke(self):
        # A BTree index with no ref lists (such as *.cix) can be dumped without
        # errors.
        builder = btree_index.BTreeBuilder(
            reference_lists=0, key_elements=2)
        builder.add_node((b'test', b'key1'), b'value')
        out_f = builder.finish()
        try:
            self.build_tree_contents([('test.btree', out_f.read())])
        finally:
            out_f.close()
        out, err = self.run_bzr('dump-btree test.btree')

    def create_sample_empty_btree_index(self):
        builder = btree_index.BTreeBuilder(
            reference_lists=1, key_elements=2)
        out_f = builder.finish()
        try:
            self.build_tree_contents([('test.btree', out_f.read())])
        finally:
            out_f.close()

    def test_dump_empty_btree_smoke(self):
        self.create_sample_empty_btree_index()
        out, err = self.run_bzr('dump-btree test.btree')
        self.assertEqualDiff("", out)

    def test_dump_empty_btree_http_smoke(self):
        self.transport_readonly_server = http_server.HttpServer
        self.create_sample_empty_btree_index()
        url = self.get_readonly_url('test.btree')
        out, err = self.run_bzr(['dump-btree', url])
        self.assertEqualDiff("", out)

    def test_dump_empty_btree_raw_smoke(self):
        self.create_sample_empty_btree_index()
        out, err = self.run_bzr('dump-btree test.btree --raw')
        self.assertEqualDiff(
            'Root node:\n'
            'B+Tree Graph Index 2\n'
            'node_ref_lists=1\n'
            'key_elements=2\n'
            'len=0\n'
            'row_lengths=\n'
            '\n'
            'Page 0\n'
            '(empty)\n',
            out)



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Implementation tests for VersionedFile-based repositories.

For more generic per-repository tests, see breezy.tests.per_repository.
"""

from breezy.tests.per_repository import (TestCaseWithRepository,
                                         all_repository_format_scenarios)


def all_repository_vf_format_scenarios():
    scenarios = []
    for test_name, scenario_info in all_repository_format_scenarios():
        format = scenario_info['repository_format']
        if format.supports_full_versioned_files:
            scenarios.append((test_name, scenario_info))
    return scenarios


def load_tests(loader, basic_tests, pattern):
    testmod_names = [
        'test_add_inventory_by_delta',
        'test_check',
        'test_check_reconcile',
        'test_find_text_key_references',
        'test__generate_text_key_index',
        'test_fetch',
        'test_fileid_involved',
        'test_merge_directive',
        'test_reconcile',
        'test_refresh_data',
        'test_repository',
        'test_write_group',
        ]
    basic_tests.addTest(loader.loadTestsFromModuleNames(
        ["{}.{}".format(__name__, tmn) for tmn in testmod_names]))
    return basic_tests



# Copyright (C) 2005, 2006, 2007, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests being able to ignore bad filetypes."""

import os
from io import StringIO

from .. import errors
from ..status import show_tree_status
from . import TestCaseWithTransport
from .features import OsFifoFeature


def verify_status(tester, tree, value):
    """Verify the output of show_tree_status"""
    tof = StringIO()
    show_tree_status(tree, to_file=tof)
    tof.seek(0)
    tester.assertEqual(value, tof.readlines())


class TestBadFiles(TestCaseWithTransport):

    def test_bad_files(self):
        """Test that bzr will ignore files it doesn't like"""
        self.requireFeature(OsFifoFeature)

        wt = self.make_branch_and_tree('.')
        b = wt.branch

        files = ['one', 'two', 'three']
        file_ids = [b'one-id', b'two-id', b'three-id']
        self.build_tree(files)
        wt.add(files, ids=file_ids)
        wt.commit("Commit one", rev_id=b"a@u-0-0")

        # We should now have a few files, lets try to
        # put some bogus stuff in the tree

        # status with nothing changed
        verify_status(self, wt, [])

        os.mkfifo('a-fifo')
        self.build_tree(['six'])

        verify_status(self, wt,
                      ['unknown:\n',
                       '  a-fifo\n',
                       '  six\n'
                       ])

        # We should raise an error if we are adding a bogus file
        self.assertRaises(errors.BadFileKindError, wt.smart_add, ['a-fifo'])

        # And the list of files shouldn't have been modified
        verify_status(self, wt,
                      ['unknown:\n',
                       '  a-fifo\n',
                       '  six\n'
                       ])

        # Make sure smart_add can handle having a bogus
        # file in the way
        wt.smart_add([])
        verify_status(self, wt,
                      ['added:\n',
                       '  six\n',
                       'unknown:\n',
                       '  a-fifo\n',
                       ])
        wt.commit("Commit four", rev_id=b"a@u-0-3")

        verify_status(self, wt,
                      ['unknown:\n',
                       '  a-fifo\n',
                       ])



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz branch."""

import os

from breezy import branch, controldir, tests
from breezy.urlutils import local_path_to_url


class TestClone(tests.TestCaseWithTransport):

    def example_dir(self, path='.', format=None):
        tree = self.make_branch_and_tree(path, format=format)
        self.build_tree_contents([(path + '/hello', b'foo')])
        tree.add('hello')
        tree.commit(message='setup')
        self.build_tree_contents([(path + '/goodbye', b'baz')])
        tree.add('goodbye')
        tree.commit(message='setup')
        return tree

    def test_clone(self):
        """Branch from one branch to another."""
        self.example_dir('a')
        self.run_bzr('clone a b')
        b = branch.Branch.open('b')
        self.run_bzr('clone a c -r 1')
        # previously was erroneously created by branching
        self.assertFalse(b._transport.has('branch-name'))
        b.controldir.open_workingtree().commit(message='foo', allow_pointless=True)

    def test_clone_no_to_location(self):
        """The to_location is derived from the source branch name."""
        os.mkdir("something")
        a = self.example_dir('something/a').branch
        self.run_bzr('clone something/a')
        b = branch.Branch.open('a')
        self.assertEqual(b.last_revision_info(), a.last_revision_info())

    def test_from_colocated(self):
        """Branch from a colocated branch into a regular branch."""
        os.mkdir('b')
        tree = self.example_dir('b/a')
        tree.controldir.create_branch(name='somecolo')
        out, err = self.run_bzr('clone %s' % local_path_to_url('b/a'))
        self.assertEqual('', out)
        self.assertEqual('Created new control directory.\n', err)
        self.assertPathExists('a')
        target = controldir.ControlDir.open('a')
        self.assertEqual(['', 'somecolo'], target.branch_names())



# Copyright (C) 2013, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz verify-signatures."""

from breezy import gpg, tests


class TestVerifySignatures(tests.TestCaseWithTransport):

    def monkey_patch_gpg(self):
        """Monkey patch the gpg signing strategy to be a loopback.

        This also registers the cleanup, so that we will revert to
        the original gpg strategy when done.
        """
        # monkey patch gpg signing mechanism
        self.overrideAttr(gpg, 'GPGStrategy', gpg.LoopbackGPGStrategy)

    def setup_tree(self, location='.'):
        wt = self.make_branch_and_tree(location)
        wt.commit("base A", allow_pointless=True, rev_id=b'A')
        wt.commit("base B", allow_pointless=True, rev_id=b'B')
        wt.commit("base C", allow_pointless=True, rev_id=b'C')
        wt.commit("base D", allow_pointless=True, rev_id=b'D',
                  committer='Alternate <alt@foo.com>')
        wt.add_parent_tree_id(b"aghost")
        wt.commit("base E", allow_pointless=True, rev_id=b'E')
        return wt

    def test_verify_signatures(self):
        wt = self.setup_tree()
        self.monkey_patch_gpg()
        self.run_bzr('sign-my-commits')
        out = self.run_bzr('verify-signatures', retcode=1)
        self.assertEqual(('4 commits with valid signatures\n'
                          '0 commits with key now expired\n'
                          '0 commits with unknown keys\n'
                          '0 commits not valid\n'
                          '1 commit not signed\n', ''), out)

    def test_verify_signatures_acceptable_key(self):
        wt = self.setup_tree()
        self.monkey_patch_gpg()
        self.run_bzr('sign-my-commits')
        out = self.run_bzr(['verify-signatures', '--acceptable-keys=foo,bar'],
                           retcode=1)
        self.assertEqual(('4 commits with valid signatures\n'
                          '0 commits with key now expired\n'
                          '0 commits with unknown keys\n'
                          '0 commits not valid\n'
                          '1 commit not signed\n', ''), out)

    def test_verify_signatures_verbose(self):
        wt = self.setup_tree()
        self.monkey_patch_gpg()
        self.run_bzr('sign-my-commits')
        out = self.run_bzr('verify-signatures --verbose', retcode=1)
        self.assertEqual(
            ('4 commits with valid signatures\n'
             '  None signed 4 commits\n'
             '0 commits with key now expired\n'
             '0 commits with unknown keys\n'
             '0 commits not valid\n'
             '1 commit not signed\n'
             '  1 commit by author Alternate <alt@foo.com>\n', ''), out)

    def test_verify_signatures_verbose_all_valid(self):
        wt = self.setup_tree()
        self.monkey_patch_gpg()
        self.run_bzr('sign-my-commits')
        self.run_bzr(['sign-my-commits', '.', 'Alternate <alt@foo.com>'])
        out = self.run_bzr('verify-signatures --verbose')
        self.assertEqual(('All commits signed with verifiable keys\n'
                          '  None signed 5 commits\n', ''), out)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for add_inventory on a repository with external references."""

from breezy import errors
from breezy.repository import WriteGroup
from breezy.tests.per_repository_reference import \
    TestCaseWithExternalReferenceRepository


class TestAddInventory(TestCaseWithExternalReferenceRepository):

    def test_add_inventory_goes_to_repo(self):
        # adding an inventory only writes to the repository add_inventory is
        # called on.
        tree = self.make_branch_and_tree('sample')
        revid = tree.commit('one')
        inv = tree.branch.repository.get_inventory(revid)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        base = self.make_repository('base')
        repo = self.make_referring('referring', base)
        with repo.lock_write(), WriteGroup(repo):
            repo.add_inventory(revid, inv, [])
        repo.lock_read()
        self.addCleanup(repo.unlock)
        inv2 = repo.get_inventory(revid)
        content1 = {file_id: inv.get_entry(file_id)
                        for file_id in inv.iter_all_ids()}
        content2 = {file_id: inv.get_entry(file_id)
                        for file_id in inv2.iter_all_ids()}
        self.assertEqual(content1, content2)
        self.assertRaises(errors.NoSuchRevision, base.get_inventory, revid)



"""
Copyright (c) 2007 Ian Cook and John Popplewell

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Date    : 11 August 2007
Version : 1.0.1
Contact : John Popplewell
Email   : john@johnnypops.demon.co.uk
Web     : http://www.johnnypops.demon.co.uk/python/
Origin  : Based on the original script by Ian Cook
          http://www.kirbyfooty.com/simplemapi.py
Comments: Works (and tested) with:
          Outlook Express, Outlook 97 and 2000,
          Eudora, Incredimail and Mozilla Thunderbird (1.5.0.2)
Thanks  : Werner F. Bruhin and Michele Petrazzo on the ctypes list.

If you have any bug-fixes, enhancements or suggestions regarding this
software, please contact me at the above email address.
"""

import os
from ctypes import *

FLAGS = c_ulong
LHANDLE = c_ulong
LPLHANDLE = POINTER(LHANDLE)

# Return codes
SUCCESS_SUCCESS = 0
MAPI_USER_ABORT = 1
MAPI_E_USER_ABORT = MAPI_USER_ABORT
MAPI_E_FAILURE = 2
MAPI_E_LOGON_FAILURE = 3
MAPI_E_LOGIN_FAILURE = MAPI_E_LOGON_FAILURE
MAPI_E_DISK_FULL = 4
MAPI_E_INSUFFICIENT_MEMORY = 5
MAPI_E_ACCESS_DENIED = 6
MAPI_E_TOO_MANY_SESSIONS = 8
MAPI_E_TOO_MANY_FILES = 9
MAPI_E_TOO_MANY_RECIPIENTS = 10
MAPI_E_ATTACHMENT_NOT_FOUND = 11
MAPI_E_ATTACHMENT_OPEN_FAILURE = 12
MAPI_E_ATTACHMENT_WRITE_FAILURE = 13
MAPI_E_UNKNOWN_RECIPIENT = 14
MAPI_E_BAD_RECIPTYPE = 15
MAPI_E_NO_MESSAGES = 16
MAPI_E_INVALID_MESSAGE = 17
MAPI_E_TEXT_TOO_LARGE = 18
MAPI_E_INVALID_SESSION = 19
MAPI_E_TYPE_NOT_SUPPORTED = 20
MAPI_E_AMBIGUOUS_RECIPIENT = 21
MAPI_E_AMBIG_RECIP = MAPI_E_AMBIGUOUS_RECIPIENT
MAPI_E_MESSAGE_IN_USE = 22
MAPI_E_NETWORK_FAILURE = 23
MAPI_E_INVALID_EDITFIELDS = 24
MAPI_E_INVALID_RECIPS = 25
MAPI_E_NOT_SUPPORTED = 26
# Recipient class
MAPI_ORIG = 0
MAPI_TO = 1
# Send flags
MAPI_LOGON_UI = 1
MAPI_DIALOG = 8


class MapiRecipDesc(Structure):
    _fields_ = [
        ('ulReserved', c_ulong),
        ('ulRecipClass', c_ulong),
        ('lpszName', c_char_p),
        ('lpszAddress', c_char_p),
        ('ulEIDSize', c_ulong),
        ('lpEntryID', c_void_p),
    ]


lpMapiRecipDesc = POINTER(MapiRecipDesc)
lppMapiRecipDesc = POINTER(lpMapiRecipDesc)


class MapiFileDesc(Structure):
    _fields_ = [
        ('ulReserved', c_ulong),
        ('flFlags', c_ulong),
        ('nPosition', c_ulong),
        ('lpszPathName', c_char_p),
        ('lpszFileName', c_char_p),
        ('lpFileType', c_void_p),
    ]


lpMapiFileDesc = POINTER(MapiFileDesc)


class MapiMessage(Structure):
    _fields_ = [
        ('ulReserved', c_ulong),
        ('lpszSubject', c_char_p),
        ('lpszNoteText', c_char_p),
        ('lpszMessageType', c_char_p),
        ('lpszDateReceived', c_char_p),
        ('lpszConversationID', c_char_p),
        ('flFlags', FLAGS),
        ('lpOriginator', lpMapiRecipDesc),
        ('nRecipCount', c_ulong),
        ('lpRecips', lpMapiRecipDesc),
        ('nFileCount', c_ulong),
        ('lpFiles', lpMapiFileDesc),
    ]


lpMapiMessage = POINTER(MapiMessage)

MAPI = windll.mapi32  # type: ignore
MAPISendMail = MAPI.MAPISendMail
MAPISendMail.restype = c_ulong
MAPISendMail.argtypes = (LHANDLE, c_ulong, lpMapiMessage, FLAGS, c_ulong)

MAPIResolveName = MAPI.MAPIResolveName
MAPIResolveName.restype = c_ulong
MAPIResolveName.argtypes = (
    LHANDLE, c_ulong, c_char_p, FLAGS, c_ulong, lppMapiRecipDesc)

MAPIFreeBuffer = MAPI.MAPIFreeBuffer
MAPIFreeBuffer.restype = c_ulong
MAPIFreeBuffer.argtypes = (c_void_p, )

MAPILogon = MAPI.MAPILogon
MAPILogon.restype = c_ulong
MAPILogon.argtypes = (LHANDLE, c_char_p, c_char_p, FLAGS, c_ulong, LPLHANDLE)

MAPILogoff = MAPI.MAPILogoff
MAPILogoff.restype = c_ulong
MAPILogoff.argtypes = (LHANDLE, c_ulong, FLAGS, c_ulong)


class MAPIError(WindowsError):  # type: ignore

    def __init__(self, code):
        WindowsError.__init__(self)  # type: ignore
        self.code = code

    def __str__(self):
        return 'MAPI error %d' % (self.code,)


def _logon(profileName=None, password=None):
    pSession = LHANDLE()
    rc = MAPILogon(0, profileName, password, MAPI_LOGON_UI, 0, byref(pSession))
    if rc != SUCCESS_SUCCESS:
        raise MAPIError(rc)
    return pSession


def _logoff(session):
    rc = MAPILogoff(session, 0, 0, 0)
    if rc != SUCCESS_SUCCESS:
        raise MAPIError(rc)


def _resolveName(session, name):
    pRecipDesc = lpMapiRecipDesc()
    rc = MAPIResolveName(session, 0, name, 0, 0, byref(pRecipDesc))
    if rc != SUCCESS_SUCCESS:
        raise MAPIError(rc)
    rd = pRecipDesc.contents
    name, address = rd.lpszName, rd.lpszAddress
    rc = MAPIFreeBuffer(pRecipDesc)
    if rc != SUCCESS_SUCCESS:
        raise MAPIError(rc)
    return name, address


def _sendMail(session, recipient, subject, body, attach):
    nFileCount = len(attach)
    if attach:
        MapiFileDesc_A = MapiFileDesc * len(attach)
        fda = MapiFileDesc_A()
        for fd, fa in zip(fda, attach):
            fd.ulReserved = 0
            fd.flFlags = 0
            fd.nPosition = -1
            fd.lpszPathName = fa
            fd.lpszFileName = None
            fd.lpFileType = None
        lpFiles = fda
    else:
        lpFiles = lpMapiFileDesc()

    RecipWork = recipient.split(';')
    RecipCnt = len(RecipWork)
    MapiRecipDesc_A = MapiRecipDesc * len(RecipWork)
    rda = MapiRecipDesc_A()
    for rd, ra in zip(rda, RecipWork):
        rd.ulReserved = 0
        rd.ulRecipClass = MAPI_TO
        try:
            rd.lpszName, rd.lpszAddress = _resolveName(session, ra)
        except OSError:  # type: ignore
            # work-round for Mozilla Thunderbird
            rd.lpszName, rd.lpszAddress = None, ra
        rd.ulEIDSize = 0
        rd.lpEntryID = None
    recip = rda

    msg = MapiMessage(0, subject, body, None, None, None, 0, lpMapiRecipDesc(),
                      RecipCnt, recip,
                      nFileCount, lpFiles)

    rc = MAPISendMail(session, 0, byref(msg), MAPI_DIALOG, 0)
    if rc != SUCCESS_SUCCESS:
        raise MAPIError(rc)


def SendMail(recipient, subject="", body="", attachfiles=""):
    """Post an e-mail message using Simple MAPI

    recipient - string: address to send to (multiple addresses separated with a semicolon)
    subject   - string: subject header
    body      - string: message text
    attach    - string: files to attach (multiple attachments separated with a semicolon)
    """

    attach = []
    AttachWork = attachfiles.split(';')
    for f in AttachWork:
        if os.path.exists(f):
            attach.append(os.path.abspath(f))

    restore = os.getcwd()
    try:
        session = _logon()
        try:
            _sendMail(session, recipient, subject, body, attach)
        finally:
            _logoff(session)
    finally:
        os.chdir(restore)


if __name__ == '__main__':
    import sys
    recipient = "test@johnnypops.demon.co.uk"
    subject = "Test Message Subject"
    body = "Hi,\r\n\r\nthis is a quick test message,\r\n\r\ncheers,\r\nJohn."
    attachment = sys.argv[0]
    SendMail(recipient, subject, body, attachment)



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for VersionedFile classes"""

from ... import errors, multiparent, tests
from .. import groupcompress, versionedfile


class Test_MPDiffGenerator(tests.TestCaseWithMemoryTransport):
    # Should this be a per vf test?

    def make_vf(self):
        t = self.get_transport('')
        factory = groupcompress.make_pack_factory(True, True, 1)
        return factory(t)

    def make_three_vf(self):
        vf = self.make_vf()
        vf.add_lines((b'one',), (), [b'first\n'])
        vf.add_lines((b'two',), [(b'one',)], [b'first\n', b'second\n'])
        vf.add_lines((b'three',), [(b'one',), (b'two',)],
                     [b'first\n', b'second\n', b'third\n'])
        return vf

    def test_finds_parents(self):
        vf = self.make_three_vf()
        gen = versionedfile._MPDiffGenerator(vf, [(b'three',)])
        needed_keys, refcount = gen._find_needed_keys()
        self.assertEqual(sorted([(b'one',), (b'two',), (b'three',)]),
                         sorted(needed_keys))
        self.assertEqual({(b'one',): 1, (b'two',): 1}, refcount)

    def test_ignores_ghost_parents(self):
        # If a parent is a ghost, it is just ignored
        vf = self.make_vf()
        vf.add_lines((b'two',), [(b'one',)], [b'first\n', b'second\n'])
        gen = versionedfile._MPDiffGenerator(vf, [(b'two',)])
        needed_keys, refcount = gen._find_needed_keys()
        self.assertEqual(sorted([(b'two',)]), sorted(needed_keys))
        # It is returned, but we don't really care as we won't extract it
        self.assertEqual({(b'one',): 1}, refcount)
        self.assertEqual([(b'one',)], sorted(gen.ghost_parents))
        self.assertEqual([], sorted(gen.present_parents))

    def test_raises_on_ghost_keys(self):
        # If the requested key is a ghost, then we have a problem
        vf = self.make_vf()
        gen = versionedfile._MPDiffGenerator(vf, [(b'one',)])
        self.assertRaises(errors.RevisionNotPresent,
                          gen._find_needed_keys)

    def test_refcount_multiple_children(self):
        vf = self.make_three_vf()
        gen = versionedfile._MPDiffGenerator(vf, [(b'two',), (b'three',)])
        needed_keys, refcount = gen._find_needed_keys()
        self.assertEqual(sorted([(b'one',), (b'two',), (b'three',)]),
                         sorted(needed_keys))
        self.assertEqual({(b'one',): 2, (b'two',): 1}, refcount)
        self.assertEqual([(b'one',)], sorted(gen.present_parents))

    def test_process_contents(self):
        vf = self.make_three_vf()
        gen = versionedfile._MPDiffGenerator(vf, [(b'two',), (b'three',)])
        gen._find_needed_keys()
        self.assertEqual({(b'two',): ((b'one',),),
                          (b'three',): ((b'one',), (b'two',))},
                         gen.parent_map)
        self.assertEqual({(b'one',): 2, (b'two',): 1}, gen.refcounts)
        self.assertEqual(sorted([(b'one',), (b'two',), (b'three',)]),
                         sorted(gen.needed_keys))
        stream = vf.get_record_stream(gen.needed_keys, 'topological', True)
        record = next(stream)
        self.assertEqual((b'one',), record.key)
        # one is not needed in the output, but it is needed by children. As
        # such, it should end up in the various caches
        gen._process_one_record(record.key, record.get_bytes_as('chunked'))
        # The chunks should be cached, the refcount untouched
        self.assertEqual({(b'one',)}, set(gen.chunks))
        self.assertEqual({(b'one',): 2, (b'two',): 1}, gen.refcounts)
        self.assertEqual(set(), set(gen.diffs))
        # Next we get 'two', which is something we output, but also needed for
        # three
        record = next(stream)
        self.assertEqual((b'two',), record.key)
        gen._process_one_record(record.key, record.get_bytes_as('chunked'))
        # Both are now cached, and the diff for two has been extracted, and
        # one's refcount has been updated. two has been removed from the
        # parent_map
        self.assertEqual({(b'one',), (b'two',)}, set(gen.chunks))
        self.assertEqual({(b'one',): 1, (b'two',): 1}, gen.refcounts)
        self.assertEqual({(b'two',)}, set(gen.diffs))
        self.assertEqual({(b'three',): ((b'one',), (b'two',))},
                         gen.parent_map)
        # Finally 'three', which allows us to remove all parents from the
        # caches
        record = next(stream)
        self.assertEqual((b'three',), record.key)
        gen._process_one_record(record.key, record.get_bytes_as('chunked'))
        # Both are now cached, and the diff for two has been extracted, and
        # one's refcount has been updated
        self.assertEqual(set(), set(gen.chunks))
        self.assertEqual({}, gen.refcounts)
        self.assertEqual({(b'two',), (b'three',)}, set(gen.diffs))

    def test_compute_diffs(self):
        vf = self.make_three_vf()
        # The content is in the order requested, even if it isn't topological
        gen = versionedfile._MPDiffGenerator(vf, [(b'two',), (b'three',),
                                                  (b'one',)])
        diffs = gen.compute_diffs()
        expected_diffs = [
            multiparent.MultiParent([multiparent.ParentText(0, 0, 0, 1),
                                     multiparent.NewText([b'second\n'])]),
            multiparent.MultiParent([multiparent.ParentText(1, 0, 0, 2),
                                     multiparent.NewText([b'third\n'])]),
            multiparent.MultiParent([multiparent.NewText([b'first\n'])]),
            ]
        self.assertEqual(expected_diffs, diffs)


class ErrorTests(tests.TestCase):

    def test_unavailable_representation(self):
        error = versionedfile.UnavailableRepresentation(
            ('key',), "mpdiff", "fulltext")
        self.assertEqualDiff("The encoding 'mpdiff' is not available for key "
                             "('key',) which is encoded as 'fulltext'.",
                             str(error))



# Copyright (C) 2007, 2008, 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility for create branches with particular contents."""

from . import commit, controldir, errors, revision


class BranchBuilder:
    r"""A BranchBuilder aids creating Branches with particular shapes.

    The expected way to use BranchBuilder is to construct a
    BranchBuilder on the transport you want your branch on, and then call
    appropriate build_ methods on it to get the shape of history you want.

    This is meant as a helper for the test suite, not as a general class for
    real data.

    For instance:

    >>> from breezy.transport.memory import MemoryTransport
    >>> builder = BranchBuilder(MemoryTransport("memory:///"))
    >>> builder.start_series()
    >>> builder.build_snapshot(None, [
    ...     ('add', ('', b'root-id', 'directory', '')),
    ...     ('add', ('filename', b'f-id', 'file', b'content\n'))],
    ...     revision_id=b'rev-id')
    b'rev-id'
    >>> builder.build_snapshot([b'rev-id'],
    ...     [('modify', ('filename', b'new-content\n'))],
    ...     revision_id=b'rev2-id')
    b'rev2-id'
    >>> builder.finish_series()
    >>> branch = builder.get_branch()

    :ivar _tree: This is a private member which is not meant to be modified by
        users of this class. While a 'series' is in progress, it should hold a
        MemoryTree with the contents of the last commit (ready to be modified
        by the next build_snapshot command) with a held write lock. Outside of
        a series in progress, it should be None.
    """

    def __init__(self, transport=None, format=None, branch=None):
        """Construct a BranchBuilder on transport.

        :param transport: The transport the branch should be created on.
            If the path of the transport does not exist but its parent does
            it will be created.
        :param format: Either a BzrDirFormat, or the name of a format in the
            controldir format registry for the branch to be built.
        :param branch: An already constructed branch to use.  This param is
            mutually exclusive with the transport and format params.
        """
        if branch is not None:
            if format is not None:
                raise AssertionError(
                    "branch and format kwargs are mutually exclusive")
            if transport is not None:
                raise AssertionError(
                    "branch and transport kwargs are mutually exclusive")
            self._branch = branch
        else:
            if not transport.has('.'):
                transport.mkdir('.')
            if format is None:
                format = 'default'
            if isinstance(format, str):
                format = controldir.format_registry.make_controldir(format)
            self._branch = controldir.ControlDir.create_branch_convenience(
                transport.base, format=format, force_new_tree=False)
        self._tree = None

    def build_commit(self, parent_ids=None, allow_leftmost_as_ghost=False,
                     **commit_kwargs):
        """Build a commit on the branch.

        This makes a commit with no real file content for when you only want
        to look at the revision graph structure.

        :param commit_kwargs: Arguments to pass through to commit, such as
             timestamp.
        """
        if parent_ids is not None:
            if len(parent_ids) == 0:
                base_id = revision.NULL_REVISION
            else:
                base_id = parent_ids[0]
            if base_id != self._branch.last_revision():
                self._move_branch_pointer(
                    base_id, allow_leftmost_as_ghost=allow_leftmost_as_ghost)
        tree = self._branch.create_memorytree()
        with tree.lock_write():
            if parent_ids is not None:
                tree.set_parent_ids(
                    parent_ids,
                    allow_leftmost_as_ghost=allow_leftmost_as_ghost)
            tree.add('')
            return self._do_commit(tree, **commit_kwargs)

    def _do_commit(self, tree, message=None, message_callback=None, **kwargs):
        reporter = commit.NullCommitReporter()
        if message is None and message_callback is None:
            message = 'commit %d' % (self._branch.revno() + 1,)
        return tree.commit(message, message_callback=message_callback,
                           reporter=reporter, **kwargs)

    def _move_branch_pointer(self, new_revision_id,
                             allow_leftmost_as_ghost=False):
        """Point self._branch to a different revision id."""
        with self._branch.lock_write():
            # We don't seem to have a simple set_last_revision(), so we
            # implement it here.
            cur_revno, cur_revision_id = self._branch.last_revision_info()
            try:
                g = self._branch.repository.get_graph()
                new_revno = g.find_distance_to_null(
                    new_revision_id, [(cur_revision_id, cur_revno)])
                self._branch.set_last_revision_info(new_revno, new_revision_id)
            except errors.GhostRevisionsHaveNoRevno:
                if not allow_leftmost_as_ghost:
                    raise
                new_revno = 1
        if self._tree is not None:
            # We are currently processing a series, but when switching branch
            # pointers, it is easiest to just create a new memory tree.
            # That way we are sure to have the right files-on-disk
            # We are cheating a little bit here, and locking the new tree
            # before the old tree is unlocked. But that way the branch stays
            # locked throughout.
            new_tree = self._branch.create_memorytree()
            new_tree.lock_write()
            self._tree.unlock()
            self._tree = new_tree

    def start_series(self):
        """We will be creating a series of commits.

        This allows us to hold open the locks while we are processing.

        Make sure to call 'finish_series' when you are done.
        """
        if self._tree is not None:
            raise AssertionError('You cannot start a new series while a'
                                 ' series is already going.')
        self._tree = self._branch.create_memorytree()
        self._tree.lock_write()

    def finish_series(self):
        """Call this after start_series to unlock the various objects."""
        self._tree.unlock()
        self._tree = None

    def build_snapshot(self, parent_ids, actions, message=None, timestamp=None,
                       allow_leftmost_as_ghost=False, committer=None,
                       timezone=None, message_callback=None, revision_id=None):
        """Build a commit, shaped in a specific way.

        Most of the actions are self-explanatory.  'flush' is special action to
        break a series of actions into discrete steps so that complex changes
        (such as unversioning a file-id and re-adding it with a different kind)
        can be expressed in a way that will clearly work.

        :param parent_ids: A list of parent_ids to use for the commit.
            It can be None, which indicates to use the last commit.
        :param actions: A list of actions to perform. Supported actions are:
            ('add', ('path', b'file-id', 'kind', b'content' or None))
            ('modify', ('path', b'new-content'))
            ('unversion', 'path')
            ('rename', ('orig-path', 'new-path'))
            ('flush', None)
        :param message: An optional commit message, if not supplied, a default
            commit message will be written.
        :param message_callback: A message callback to use for the commit, as
            per mutabletree.commit.
        :param timestamp: If non-None, set the timestamp of the commit to this
            value.
        :param timezone: An optional timezone for timestamp.
        :param committer: An optional username to use for commit
        :param allow_leftmost_as_ghost: True if the leftmost parent should be
            permitted to be a ghost.
        :param revision_id: The handle for the new commit, can be None
        :return: The revision_id of the new commit
        """
        if parent_ids is not None:
            if len(parent_ids) == 0:
                base_id = revision.NULL_REVISION
            else:
                base_id = parent_ids[0]
            if base_id != self._branch.last_revision():
                self._move_branch_pointer(
                    base_id, allow_leftmost_as_ghost=allow_leftmost_as_ghost)

        if self._tree is not None:
            tree = self._tree
        else:
            tree = self._branch.create_memorytree()
        with tree.lock_write():
            if parent_ids is not None:
                tree.set_parent_ids(
                    parent_ids,
                    allow_leftmost_as_ghost=allow_leftmost_as_ghost)
            # Unfortunately, MemoryTree.add(directory) just creates an
            # inventory entry. And the only public function to create a
            # directory is MemoryTree.mkdir() which creates the directory, but
            # also always adds it. So we have to use a multi-pass setup.
            pending = _PendingActions()
            for action, info in actions:
                if action == 'add':
                    path, file_id, kind, content = info
                    if kind == 'directory':
                        pending.to_add_directories.append((path, file_id))
                    else:
                        pending.to_add_files.append(path)
                        pending.to_add_file_ids.append(file_id)
                        pending.to_add_kinds.append(kind)
                        if content is not None:
                            pending.new_contents[path] = content
                elif action == 'modify':
                    path, content = info
                    pending.new_contents[path] = content
                elif action == 'unversion':
                    pending.to_unversion_paths.add(info)
                elif action == 'rename':
                    from_relpath, to_relpath = info
                    pending.to_rename.append((from_relpath, to_relpath))
                elif action == 'flush':
                    self._flush_pending(tree, pending)
                    pending = _PendingActions()
                else:
                    raise ValueError('Unknown build action: "{}"'.format(action))
            self._flush_pending(tree, pending)
            return self._do_commit(
                tree, message=message, rev_id=revision_id,
                timestamp=timestamp, timezone=timezone, committer=committer,
                message_callback=message_callback)

    def _flush_pending(self, tree, pending):
        """Flush the pending actions in 'pending', i.e. apply them to tree."""
        for path, file_id in pending.to_add_directories:
            if path == '':
                if tree.has_filename(path) \
                        and path in pending.to_unversion_paths:
                    # We're overwriting this path, no need to unversion
                    pending.to_unversion_paths.discard(path)
                # Special case, because the path already exists
                if file_id is not None:
                    tree.add([path], ['directory'], ids=[file_id])
                else:
                    tree.add([path], ['directory'])
            else:
                if file_id is not None:
                    tree.mkdir(path, file_id)
                else:
                    tree.mkdir(path)
        for from_relpath, to_relpath in pending.to_rename:
            tree.rename_one(from_relpath, to_relpath)
        if pending.to_unversion_paths:
            tree.unversion(pending.to_unversion_paths)
        if tree.supports_file_ids:
            tree.add(pending.to_add_files,
                     pending.to_add_kinds, pending.to_add_file_ids)
        else:
            tree.add(pending.to_add_files, pending.to_add_kinds)
        for path, content in pending.new_contents.items():
            tree.put_file_bytes_non_atomic(path, content)

    def get_branch(self):
        """Return the branch created by the builder."""
        return self._branch


class _PendingActions:
    """Pending actions for build_snapshot to take.

    This is just a simple class to hold a bunch of the intermediate state of
    build_snapshot in single object.
    """

    def __init__(self):
        self.to_add_directories = []
        self.to_add_files = []
        self.to_add_file_ids = []
        self.to_add_kinds = []
        self.new_contents = {}
        self.to_unversion_paths = set()
        self.to_rename = []



# Copyright (C) 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A collection of commonly used 'Features' to optionally run tests.
"""

import importlib
import os
import stat
import subprocess
import sys
import tempfile
import warnings

from .. import osutils, symbol_versioning


class Feature:
    """An operating system Feature."""

    def __init__(self):
        self._available = None

    def available(self):
        """Is the feature available?

        :return: True if the feature is available.
        """
        if self._available is None:
            self._available = self._probe()
        return self._available

    def _probe(self):
        """Implement this method in concrete features.

        :return: True if the feature is available.
        """
        raise NotImplementedError

    def __str__(self):
        if getattr(self, 'feature_name', None):
            return self.feature_name()
        return self.__class__.__name__


class SymlinkFeature(Feature):
    """Whether symlinks can be created by the current user."""

    def __init__(self, path):
        super().__init__()
        self.path = path

    def _probe(self):
        return osutils.supports_symlinks(self.path)

    def feature_name(self):
        return 'symlinks'


class HardlinkFeature(Feature):

    def __init__(self, path):
        super().__init__()
        self.path = path

    def _probe(self):
        return osutils.supports_hardlinks(self.path)

    def feature_name(self):
        return 'hardlinks'


class _OsFifoFeature(Feature):

    def _probe(self):
        return getattr(os, 'mkfifo', None)

    def feature_name(self):
        return 'filesystem fifos'


OsFifoFeature = _OsFifoFeature()


class _UnicodeFilenameFeature(Feature):
    """Does the filesystem support Unicode filenames?"""

    def _probe(self):
        try:
            # Check for character combinations unlikely to be covered by any
            # single non-unicode encoding. We use the characters
            # - greek small letter alpha (U+03B1) and
            # - braille pattern dots-123456 (U+283F).
            os.stat('\u03b1\u283f')
        except UnicodeEncodeError:
            return False
        except OSError:
            # The filesystem allows the Unicode filename but the file doesn't
            # exist.
            return True
        else:
            # The filesystem allows the Unicode filename and the file exists,
            # for some reason.
            return True


UnicodeFilenameFeature = _UnicodeFilenameFeature()


class _CompatabilityThunkFeature(Feature):
    """This feature is just a thunk to another feature.

    It issues a deprecation warning if it is accessed, to let you know that you
    should really use a different feature.
    """

    def __init__(self, dep_version, module, name,
                 replacement_name, replacement_module=None):
        super().__init__()
        self._module = module
        if replacement_module is None:
            replacement_module = module
        self._replacement_module = replacement_module
        self._name = name
        self._replacement_name = replacement_name
        self._dep_version = dep_version
        self._feature = None

    def _ensure(self):
        if self._feature is None:
            from breezy import pyutils
            depr_msg = self._dep_version % ('%s.%s'
                                            % (self._module, self._name))
            use_msg = ' Use {}.{} instead.'.format(self._replacement_module,
                                               self._replacement_name)
            symbol_versioning.warn(depr_msg + use_msg, DeprecationWarning,
                                   stacklevel=5)
            # Import the new feature and use it as a replacement for the
            # deprecated one.
            self._feature = pyutils.get_named_object(
                self._replacement_module, self._replacement_name)

    def _probe(self):
        self._ensure()
        return self._feature._probe()


class ModuleAvailableFeature(Feature):
    """This is a feature than describes a module we want to be available.

    Declare the name of the module in __init__(), and then after probing, the
    module will be available as 'self.module'.

    :ivar module: The module if it is available, else None.
    """

    def __init__(self, module_name, ignore_warnings=None):
        super().__init__()
        self.module_name = module_name
        if ignore_warnings is None:
            ignore_warnings = ()
        self.ignore_warnings = ignore_warnings

    def _probe(self):
        sentinel = object()
        module = sys.modules.get(self.module_name, sentinel)
        if module is sentinel:
            with warnings.catch_warnings():
                for warning_category in self.ignore_warnings:
                    warnings.simplefilter('ignore', warning_category)
                try:
                    self._module = importlib.import_module(self.module_name)
                except ImportError:
                    return False
                return True
        else:
            self._module = module
            return True

    @property
    def module(self):
        if self.available():
            return self._module
        return None

    def feature_name(self):
        return self.module_name


class PluginLoadedFeature(Feature):
    """Check whether a plugin with specific name is loaded.

    This is different from ModuleAvailableFeature, because
    plugins can be available but explicitly disabled
    (e.g. through BRZ_DISABLE_PLUGINS=blah).

    :ivar plugin_name: The name of the plugin
    """

    def __init__(self, plugin_name):
        super().__init__()
        self.plugin_name = plugin_name

    def _probe(self):
        from breezy.plugin import get_loaded_plugin
        return (get_loaded_plugin(self.plugin_name) is not None)

    @property
    def plugin(self):
        from breezy.plugin import get_loaded_plugin
        return get_loaded_plugin(self.plugin_name)

    def feature_name(self):
        return '%s plugin' % self.plugin_name


class _HTTPSServerFeature(Feature):
    """Some tests want an https Server, check if one is available.
    """

    def _probe(self):
        try:
            import ssl  # noqa: F401
            return True
        except ModuleNotFoundError:
            return False

    def feature_name(self):
        return 'HTTPSServer'


HTTPSServerFeature = _HTTPSServerFeature()


class _ByteStringNamedFilesystem(Feature):
    """Is the filesystem based on bytes?"""

    def _probe(self):
        if os.name == "posix":
            return True
        return False


ByteStringNamedFilesystem = _ByteStringNamedFilesystem()


class _UTF8Filesystem(Feature):
    """Is the filesystem UTF-8?"""

    def _probe(self):
        if sys.getfilesystemencoding().upper() in ('UTF-8', 'UTF8'):
            return True
        return False


UTF8Filesystem = _UTF8Filesystem()


class _BreakinFeature(Feature):
    """Does this platform support the breakin feature?"""

    def _probe(self):
        from breezy import breakin
        if breakin.determine_signal() is None:
            return False
        if sys.platform == 'win32':
            # Windows doesn't have os.kill, and we catch the SIGBREAK signal.
            # We trigger SIGBREAK via a Console api so we need ctypes to
            # access the function
            try:
                import ctypes  # noqa: F401
            except OSError:
                return False
        return True

    def feature_name(self):
        return "SIGQUIT or SIGBREAK w/ctypes on win32"


BreakinFeature = _BreakinFeature()


class _CaseInsCasePresFilenameFeature(Feature):
    """Is the file-system case insensitive, but case-preserving?"""

    def _probe(self):
        fileno, name = tempfile.mkstemp(prefix='MixedCase')
        try:
            # first check truly case-preserving for created files, then check
            # case insensitive when opening existing files.
            name = osutils.normpath(name)
            base, rel = osutils.split(name)
            found_rel = osutils.canonical_relpath(base, name)
            return (found_rel == rel and
                    os.path.isfile(name.upper()) and
                    os.path.isfile(name.lower()))
        finally:
            os.close(fileno)
            os.remove(name)

    def feature_name(self):
        return "case-insensitive case-preserving filesystem"


CaseInsCasePresFilenameFeature = _CaseInsCasePresFilenameFeature()


class _CaseInsensitiveFilesystemFeature(Feature):
    """Check if underlying filesystem is case-insensitive but *not* case
    preserving.
    """
    # Note that on Windows, Cygwin, MacOS etc, the file-systems are far
    # more likely to be case preserving, so this case is rare.

    def _probe(self):
        if CaseInsCasePresFilenameFeature.available():
            return False

        from breezy import tests

        if tests.TestCaseWithMemoryTransport.TEST_ROOT is None:
            root = tempfile.mkdtemp(prefix='testbzr-', suffix='.tmp')
            tests.TestCaseWithMemoryTransport.TEST_ROOT = root
        else:
            root = tests.TestCaseWithMemoryTransport.TEST_ROOT
        tdir = tempfile.mkdtemp(prefix='case-sensitive-probe-', suffix='',
                                dir=root)
        name_a = osutils.pathjoin(tdir, 'a')
        name_A = osutils.pathjoin(tdir, 'A')
        os.mkdir(name_a)
        result = osutils.isdir(name_A)
        tests._rmtree_temp_dir(tdir)
        return result

    def feature_name(self):
        return 'case-insensitive filesystem'


CaseInsensitiveFilesystemFeature = _CaseInsensitiveFilesystemFeature()


class _CaseSensitiveFilesystemFeature(Feature):

    def _probe(self):
        if CaseInsCasePresFilenameFeature.available():
            return False
        elif CaseInsensitiveFilesystemFeature.available():
            return False
        else:
            return True

    def feature_name(self):
        return 'case-sensitive filesystem'


# new coding style is for feature instances to be lowercase
case_sensitive_filesystem_feature = _CaseSensitiveFilesystemFeature()


class _NotRunningAsRoot(Feature):

    def _probe(self):
        try:
            uid = os.getuid()
        except AttributeError:
            # If there is no uid, chances are there is no root either
            return True
        return uid != 0

    def feature_name(self):
        return 'Not running as root'


not_running_as_root = _NotRunningAsRoot()

# Apport uses deprecated imp module on python3.
apport = ModuleAvailableFeature(
    'apport.report',
    ignore_warnings=[DeprecationWarning, PendingDeprecationWarning])
gpg = ModuleAvailableFeature('gpg')
lzma = ModuleAvailableFeature('lzma')
meliae = ModuleAvailableFeature('meliae.scanner')
paramiko = ModuleAvailableFeature('paramiko')
pywintypes = ModuleAvailableFeature('pywintypes')
subunit = ModuleAvailableFeature('subunit')
testtools = ModuleAvailableFeature('testtools')
flake8 = ModuleAvailableFeature('flake8.api.legacy')

lsprof_feature = ModuleAvailableFeature('breezy.lsprof')

pyinotify = ModuleAvailableFeature('pyinotify')


class _BackslashDirSeparatorFeature(Feature):

    def _probe(self):
        try:
            os.lstat(os.getcwd() + '\\')
        except OSError:
            return False
        else:
            return True

    def feature_name(self):
        return "Filesystem treats '\\' as a directory separator."


backslashdir_feature = _BackslashDirSeparatorFeature()


class _ChownFeature(Feature):
    """os.chown is supported"""

    def _probe(self):
        return os.name == 'posix' and hasattr(os, 'chown')


chown_feature = _ChownFeature()


class ExecutableFeature(Feature):
    """Feature testing whether an executable of a given name is on the PATH."""

    def __init__(self, name):
        super().__init__()
        self.name = name
        self._path = None

    @property
    def path(self):
        # This is a property, so accessing path ensures _probe was called
        self.available()
        return self._path

    def _probe(self):
        self._path = osutils.find_executable_on_path(self.name)
        return self._path is not None

    def feature_name(self):
        return '%s executable' % self.name


bash_feature = ExecutableFeature('bash')
diff_feature = ExecutableFeature('diff')
sed_feature = ExecutableFeature('sed')
msgmerge_feature = ExecutableFeature('msgmerge')


class _PosixPermissionsFeature(Feature):

    def _probe(self):
        def has_perms():
            # Create temporary file and check if specified perms are
            # maintained.
            write_perms = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
            f = tempfile.mkstemp(prefix='bzr_perms_chk_')
            fd, name = f
            os.close(fd)
            osutils.chmod_if_possible(name, write_perms)

            read_perms = os.stat(name).st_mode & 0o777
            os.unlink(name)
            return (write_perms == read_perms)

        return (os.name == 'posix') and has_perms()

    def feature_name(self):
        return 'POSIX permissions support'


posix_permissions_feature = _PosixPermissionsFeature()


class _StraceFeature(Feature):

    def _probe(self):
        try:
            proc = subprocess.Popen(['strace'],
                                    stderr=subprocess.PIPE,
                                    stdout=subprocess.PIPE)
            proc.communicate()
            return True
        except OSError as e:
            import errno
            if e.errno == errno.ENOENT:
                # strace is not installed
                return False
            else:
                raise

    def feature_name(self):
        return 'strace'


strace_feature = _StraceFeature()


class _AttribFeature(Feature):

    def _probe(self):
        if (sys.platform not in ('cygwin', 'win32')):
            return False
        try:
            proc = subprocess.Popen(['attrib', '.'], stdout=subprocess.PIPE)
        except OSError:
            return False
        return (0 == proc.wait())

    def feature_name(self):
        return 'attrib Windows command-line tool'


AttribFeature = _AttribFeature()


class Win32Feature(Feature):
    """Feature testing whether we're running selftest on Windows
    or Windows-like platform.
    """

    def _probe(self):
        return sys.platform == 'win32'

    def feature_name(self):
        return "win32 platform"


win32_feature = Win32Feature()


class _BackslashFilenameFeature(Feature):
    """Does the filesystem support backslashes in filenames?"""

    def _probe(self):

        try:
            fileno, name = tempfile.mkstemp(prefix='bzr\\prefix')
        except OSError:
            return False
        else:
            try:
                os.stat(name)
            except OSError:
                # mkstemp succeeded but the file wasn't actually created
                return False
            os.close(fileno)
            os.remove(name)
            return True


BackslashFilenameFeature = _BackslashFilenameFeature()


class PathFeature(Feature):
    """Feature testing whether a particular path exists."""

    def __init__(self, path):
        super().__init__()
        self.path = path

    def _probe(self):
        return os.path.exists(self.path)

    def feature_name(self):
        return "%s exists" % self.path



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Author: Martin Pool <mbp@canonical.com>

"""Store and retrieve weaves in files.

There is one format marker followed by a blank line, followed by a
series of version headers, followed by the weave itself.

Each version marker has

 'i'   parent version indexes
 '1'   SHA-1 of text
 'n'   name

The inclusions do not need to list versions included by a parent.

The weave is bracketed by 'w' and 'W' lines, and includes the '{}[]'
processing instructions.  Lines of text are prefixed by '.' if the
line contains a newline, or ',' if not.
"""

# TODO: When extracting a single version it'd be enough to just pass
# an iterator returning the weave lines...  We don't really need to
# deserialize it into memory.

FORMAT_1 = b'# bzr weave file v5\n'


def write_weave(weave, f, format=None):
    if format is None or format == 1:
        return write_weave_v5(weave, f)
    else:
        raise ValueError("unknown weave format %r" % format)


def write_weave_v5(weave, f):
    """Write weave to file f."""
    f.write(FORMAT_1)

    for version, included in enumerate(weave._parents):
        if included:
            # mininc = weave.minimal_parents(version)
            mininc = included
            f.write(b'i ')
            f.write(b' '.join(b'%d' % i for i in mininc))
            f.write(b'\n')
        else:
            f.write(b'i\n')
        f.write(b'1 ' + weave._sha1s[version] + b'\n')
        f.write(b'n ' + weave._names[version] + b'\n')
        f.write(b'\n')

    f.write(b'w\n')

    for l in weave._weave:
        if isinstance(l, tuple):
            if l[0] == b'}':
                f.write(b'}\n')
            else:
                f.write(l[0] + b' %d\n' % l[1])
        else:  # text line
            if not l:
                f.write(b', \n')
            elif l.endswith(b'\n'):
                f.write(b'. ' + l)
            else:
                f.write(b', ' + l + b'\n')

    f.write(b'W\n')


def read_weave(f):
    # FIXME: detect the weave type and dispatch
    from .weave import Weave
    w = Weave(getattr(f, 'name', None))
    _read_weave_v5(f, w)
    return w


def _read_weave_v5(f, w):
    """Private helper routine to read a weave format 5 file into memory.

    This is only to be used by read_weave and WeaveFile.__init__.
    """
    #  200   0   2075.5080   1084.0360   breezy.weavefile:104(_read_weave_v5)
    # +60412 0    366.5900    366.5900   +<method 'readline' of 'file' objects>
    # +59982 0    320.5280    320.5280   +<method 'startswith' of 'str' objects>
    # +59363 0    297.8080    297.8080   +<method 'append' of 'list' objects>
    # replace readline call with iter over all lines ->
    # safe because we already suck on memory.
    #  200   0   1492.7170    802.6220   breezy.weavefile:104(_read_weave_v5)
    # +59982 0    329.9100    329.9100   +<method 'startswith' of 'str' objects>
    # +59363 0    320.2980    320.2980   +<method 'append' of 'list' objects>
    # replaced startswith with slice lookups:
    #  200   0    851.7250    501.1120   breezy.weavefile:104(_read_weave_v5)
    # +59363 0    311.8780    311.8780   +<method 'append' of 'list' objects>
    # +200   0     30.2500     30.2500   +<method 'readlines' of 'file' objects>

    from .weave import WeaveFormatError

    try:
        lines = iter(f.readlines())
    finally:
        f.close()

    try:
        l = next(lines)
    except StopIteration:
        raise WeaveFormatError('invalid weave file: no header')

    if l != FORMAT_1:
        raise WeaveFormatError('invalid weave file header: %r' % l)

    ver = 0
    # read weave header.
    while True:
        l = next(lines)
        if l[0:1] == b'i':
            if len(l) > 2:
                w._parents.append(list(map(int, l[2:].split(b' '))))
            else:
                w._parents.append([])
            l = next(lines)[:-1]
            w._sha1s.append(l[2:])
            l = next(lines)
            name = l[2:-1]
            w._names.append(name)
            w._name_map[name] = ver
            l = next(lines)
            ver += 1
        elif l == b'w\n':
            break
        else:
            raise WeaveFormatError('unexpected line %r' % l)

    # read weave body
    while True:
        l = next(lines)
        if l == b'W\n':
            break
        elif b'. ' == l[0:2]:
            w._weave.append(l[2:])  # include newline
        elif b', ' == l[0:2]:
            w._weave.append(l[2:-1])        # exclude newline
        elif l == b'}\n':
            w._weave.append((b'}', None))
        else:
            w._weave.append((l[0:1], int(l[2:].decode('ascii'))))
    return w
